/* erode.c - erosion and cresting functions
 *
 * Copyright (C) 2001-2005 Patrice St-Gelais
 *         patrstg@users.sourceforge.net
 *         www.oricom.ca/patrice.st-gelais
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "hf.h"
#include "erode.h"

#ifndef MAX_CACHE
#define MAX_CACHE 32768
#endif

static gdouble **cache;
static gboolean if_cache_init=FALSE;

rain_erosion_struct *rain_erosion_struct_new (gint drops, gint threshold) {
	rain_erosion_struct *res;
	res = (rain_erosion_struct *) malloc(sizeof(rain_erosion_struct));
	res->drops = drops;
	res->threshold = threshold;
	
	return res;
}

void rain_erosion_struct_free (rain_erosion_struct *res) {
	if (res)
		free(res);
}

gravity_struct *gravity_struct_new (gint steps, gint slope_threshold) {
	gravity_struct *ges;
	ges = (gravity_struct *) malloc(sizeof(gravity_struct));
	ges->steps = steps;
	ges->threshold = slope_threshold;
	return ges;
}

void gravity_struct_free (gravity_struct *ges){
	if (ges)
		free(ges);
}

oriented_gravity_struct *oriented_gravity_struct_new (gint steps, gint threshold, gint direction) {
	oriented_gravity_struct *ogs;
	ogs = (oriented_gravity_struct *) malloc(sizeof(oriented_gravity_struct));
	ogs->steps = steps;
	ogs->threshold = threshold;
	ogs->direction = direction;
	return ogs;
}

void oriented_gravity_struct_free (oriented_gravity_struct *ogs) {
	if (ogs)
		free(ogs);
}

whimsical_erosion_struct *whimsical_erosion_struct_new (gint radius, gint smooth_ribs, gboolean auto_contrast) {
	whimsical_erosion_struct *wes;
	wes = (whimsical_erosion_struct *) malloc(sizeof(whimsical_erosion_struct));
	wes->radius = radius;
	wes->smooth_ribs = smooth_ribs;
	wes->auto_contrast = auto_contrast;
	return wes;
}

void whimsical_erosion_struct_free (whimsical_erosion_struct *wes){
	if (wes)
		free(wes);
}

craters_erosion_struct *craters_erosion_struct_new (gint qty, gint display_interval, gint peak_threshold, gint slope_threshold) {
	craters_erosion_struct *ces;
	ces = (craters_erosion_struct *) malloc(sizeof(craters_erosion_struct));
	ces->qty = qty;
	ces->display_interval = display_interval;
	ces->peak_threshold = peak_threshold;
	ces->slope_threshold = slope_threshold;
	ces->crater_str = crater_struct_new();
	return ces;
}

void craters_erosion_struct_free (craters_erosion_struct *ces) { 
	if (!ces)
		return;
	if (ces->crater_str)
		crater_struct_free(ces->crater_str);
	free(ces);
}

void hf_eroded_ribs (hf_type *hf_in, hf_type *hf_out, gint max) {
//	Build a "ribbed" (eroded) template by taking the minimum
//	absolute value of the difference between each pixel and its neighborus
//	Neighbours are numbered from 1 to 8, clockwise, 1 being (1,0)
//	The default for this version of the program is to wrap
	static gint z1, z2, z3, z4, z5, z6, z7, z8;
	hf_type value;
	static gint x,y, i, val;
	for (y=0; y<max; y++)
		for (x=0; x<max; x++) {
			i = VECTORIZE(x,y,max);
			value = *(hf_in + i);
			z1 =  *(hf_in+VECTORIZE(WRAP2(x+1,max),y,max))  ;
			z2 = *(hf_in+VECTORIZE(WRAP2(x+1,max),
				WRAP2(y+1,max),max)) ;
			z3 = *(hf_in+VECTORIZE(x,WRAP2(y+1,max),max));
			z4 = *(hf_in+VECTORIZE(WRAP2(x-1,max),WRAP2(y+1,max),max));
			z5 =  *(hf_in+VECTORIZE(WRAP2(x-1,max),y,max)) ;
			z6 = *(hf_in+VECTORIZE(WRAP2(x-1,max),
				WRAP2(y-1,max),max)) ;
			z7 = *(hf_in+VECTORIZE(x,WRAP2(y-1,max),max)) ;
			z8 = *(hf_in+VECTORIZE(WRAP2(x+1,max),
				WRAP2(y-1,max),max)) ;

//			Absolute minimum (negative of)
			z1 = ABS(z1-value);
			z2 = ABS(z2-value);
			z3 = ABS(z3-value);
			z4 = ABS(z4-value);
			z5 = ABS(z5-value);
			z6 = ABS(z6-value);
			z7 = ABS(z7-value);
			z8 = ABS(z8-value);

			*(hf_out + i)  = ~ (hf_type) MIN(z1,MIN(z2,MIN(z3,MIN(z4,MIN(z5,MIN(z6,MIN(z7,z8))))))) ;
// if (!(i%512)) {
// printf("*(hf_out+i): %d; z1: %d, z2: %d, z3: %d, z4: %d, z5: %d, z6: %d, z7: %d, z8: %d\n",*(hf_out+i), z1, z2, z3, z4, z5, z6, z7, z8);
// }
		}
}

void hf_oriented_relax (hf_struct_type *hf, gint steps, gint max_slope,  gint direction) {
	// Oriented gravity relaxation (for wind functions)
	
	gint v1, v2, i, slope, x, y, iprec, step;
	hf_type value;
	
	// Direction is one of NORTH (ymax to 0), SOUTH (0 to ymax), WEST (xmax to 0), EAST (0 to xmax)
	// Note: spreading the values with a 3x3 kernel gives no significant visible difference
	slope = iget_absolute_slope(max_slope,hf->max_x);
	for (step=0; step<steps; step++) {
		for (y=0; y<hf->max_y; y++) {
			for (x=0; x<hf->max_x; x++) {
				switch (direction) {
					case NORTH: // For N and S, swap x & y
						i = VECTORIZE(x,hf->max_x-y-1,hf->max_x);
						iprec = VECTORIZE (x, WRAP2(hf->max_x-y,hf->max_x), hf->max_x);
						break;
					case SOUTH:	
						i = VECTORIZE(x,y,hf->max_x);
						iprec = VECTORIZE ( x, WRAP2(y-1,hf->max_x), hf->max_x);
						break;
					case WEST:		
						i = VECTORIZE(hf->max_x-x-1,y,hf->max_x);
						iprec = VECTORIZE ( WRAP2(hf->max_x-x,hf->max_x), y, hf->max_x);
						break;
					default: // East		
						i = VECTORIZE(x,y,hf->max_x);
						iprec = VECTORIZE ( WRAP2(x-1,hf->max_x), y, hf->max_x);
				}
				
				v1 = *(hf->hf_buf+iprec);
				v2 =  ((gint) *(hf->hf_buf+i)) + slope;
				if ( v1 > v2 ) {

					*(hf->hf_buf+iprec) -= (v1-v2) /2;
					*(hf->hf_buf+i) += (v1-v2) / 2  ;
				 
				}
			}
		}
	}
}

gboolean relax_hex (hf_type *hf, gint max_x, gint max_y, gint slope_threshold, gint wrap, gboolean even, gboolean shift_right) {

//	Gravity relaxation,on an hexagonally sampled buffer
//	Variation over cresting, more natural-like (I hope!)
//	Written 2005-02 for sand / wind processes
//	hf: buffer to transform
//	x,y: pixel to relax
//	max_x, max_y: size of the buffer (so it could work with non-square buffers!)
//	slope_threshold: maximum allowed slope, in absolute HF units (not degrees!)
//	max_steps: don't do more than this number of steps
//		The function can stop before if all slopes are lowver than max_slope
//	even, shift_right: hexagonal sampling parameters (see comments in other functions)
//	Returns: TRUE is relaxation done, FALSE if no relaxation to do
//

	static gint v0[6] = {0,0,0,0,0,0};
	static gint x,y,v[6],z1, z2, z3, z4, z5, z6, i, ii, nx, ny, ninf, shift, sumdif;
	long int value;
	gboolean test, found=FALSE;
	
	// The beginning is the same as find_all_neighbours
	// Then we take all the neighbour pixels inferior to the threshold
	// We average the difference between these pixels and the threshold
	// Then the current pixel is decreased and the "soil" is evenly
	// spread on the selected pixels

	for (x=0; x<max_x; x++) {
		for (y=0; y<max_y; y++) {
			ninf = 0;
			sumdif = 0;
			memcpy(v,v0,6*sizeof(gint));
			value = *(hf+VECTORIZE(x,y,max_x)) - slope_threshold ;
			test = y & (gint) 1; // test = TRUE if value is odd
			if ( (test && even) || ((!test) && (!even)) )  //  XOR
				shift = 1;
			else
				shift = 0;
	
			ii=VECTORIZE(WRAP2(x+1,max_x),y,max_x);
			z1 =  *(hf+ii);
			if (z1<value) {
				if (wrap || ( (x+1)<max_x ) ) {
					v[ninf] = ii;
					ninf++;
					sumdif += value-z1;
				}
			}
			
			nx = WRAP2(x+shift,max_x);
			ny = WRAP2(y+1,max_y);
			ii=VECTORIZE(nx, ny,max_x) ;
			z2 = *(hf+ii);
			if (z2<value) {
				if (wrap || (((x+shift)<max_x) && ((y+1)<max_y)) ) {
					v[ninf] = ii;
					ninf++;
					sumdif += value-z2;
				}
			}
				
			nx = WRAP2(x-1+shift,max_x);	
			ny = WRAP2(y+1,max_y) ;
			ii =VECTORIZE(nx,ny,max_x) ;
			z3 = *(hf+ii) ;
			if (z3<value) {
				if (wrap || (((x+shift)>0) && ((y+1)<max_y)) ) {
					v[ninf] = ii;
					ninf++;
					sumdif += value-z3;
				}
			}
				
			nx = WRAP2(x-1,max_x);
			ii = VECTORIZE(nx,y,max_x)  ;
			z4 = *(hf+ii);
			if (z4<value) {
				if (wrap || (x>0)) {
					v[ninf] = ii;
					ninf++;
					sumdif += value-z4;
				}
			}
			
			nx = WRAP2(x-1+shift,max_x) ;	
			ny = WRAP2(y-1,max_y) ;
			ii = VECTORIZE(nx,ny,max_x) ;
			z5 =  *(hf+ii) ;
			if (z5<value) {
				if (wrap || ((y>0) && ((x+shift)>0)) ) {
					v[ninf] = ii;
					ninf++;
					sumdif += value-z5;
				}
			}
				
			nx = WRAP2(x+shift,max_x);
			ny = WRAP2(y-1,max_y);
			ii = VECTORIZE(nx, ny,max_x) ;
			z6 = *(hf+ii);
			if (z6<value) {
				if (wrap || ((y>0) && ((x+shift)<max_x)) ) { 
					v[ninf] = ii;
					ninf++;
					sumdif += value-z6;
				}
			}
		
			if (ninf)
				found = TRUE;
			else
				continue;	

// if (y==255)
// printf("(X,Y): (%d,%d); ninf: %d; sl_max: %d; value: %d; sumdif: %d; sd/ninf+1: %d; sd/ninf*ninf+1: %d\n",x,y,ninf,slope_threshold, value, sumdif, sumdif/(ninf+1), sumdif / (ninf*(ninf+1))  );			
			
			*(hf+VECTORIZE(x,y,max_x)) -= sumdif / (ninf+1);
			
			sumdif = sumdif / (ninf*(ninf+1));
			
			for (i=0; i<ninf; i++) {
				*(hf+v[i]) += sumdif;
			}

	
		} // Y loop
		
	} // X loop
	
	return found;
}

void hf_relax_hex (hf_struct_type *hf, gint max_steps, gint max_slope) {

//	Wrapper for relax_hex - gravity erosion
	
	gint steps=0, slope;
	hf_type *tmp;
	gboolean even=TRUE;
	
	tmp = hf->hf_buf;
	hf->hf_buf = hexagonal_row_sampling (hf->hf_buf, hf->max_x, 
		hf->if_tiles, even, TRUE); 
	free(tmp);
	
	slope = iget_absolute_slope(max_slope,hf->max_x);
	
	for (steps=0; steps<max_steps; steps++) {
//	printf("******************** STEP %d **********************\n",steps);
		if (!relax_hex (hf->hf_buf, hf->max_x, hf->max_y, slope, hf->if_tiles, even, TRUE))
			break;
	}
	
//	printf("Steps: %d / %d\n",steps, max_steps);
	
	tmp = hf->hf_buf;
	hf->hf_buf = hexagonal_row_sampling (hf->hf_buf, hf->max_x, 
		hf->if_tiles, even, FALSE); 
	free(tmp);
}

gboolean find_all_neighbours_hex( gint x, gint y, gint *newx, gint *newy,
	gint max, hf_type *hf, gint threshold, 
	gboolean even, gboolean shift_right) {
	//	Find all the inferior neighbours of (x,y) in hf
	//	Randomly select one of the neighbours,
	//	given its height difference with the minimum is < to threshold
	//	Returns the result in newx and newy
	//	If (x,y) is a minimum, returns FALSE
	//	This version works on an hexagonally samplec HF (six neighbours)
	//	We need to know if even or odd rows have been shifted (even=TRUE)
	//	and if they have been shifted right or left
	//	Neighbours are counted clockwise from 1 to 6, 1 being (1,0)
	//	2, 3, 5 and 6 are being shifted on x, depending on "even" and "shift_right"
	//
	//             5 | 6
	//           4 | 0 | 1
	//             3 | 2
	//
	static gint v[6],z1, z2, z3, z4, z5, z6, i, ii, nx, ny, ninf, ok, shift;
	long int value;
	long int lower= (long int) 0xFFFF;
	gboolean test;
	ninf = 0;
	ok = 0;
	value = *(hf+VECTORIZE(x,y,max)) ;
	
	test = y & (gint) 1; // test = TRUE if value is odd
	if ( (test && even) || ((!test) && (!even)) )  //  XOR
		shift = 1;
	else
		shift = 0;
	
	ii=VECTORIZE(WRAP2(x+1,max),y,max);
	z1 =  *(hf+ii);
	if (z1<value) {
		v[ninf] = ii;
		ninf++;
	}
	lower = z1;
	
	nx = WRAP2(x+shift,max);
	ny = WRAP2(y+1,max);
	ii=VECTORIZE(nx, ny,max) ;
	z2 = *(hf+ii);
	if (z2<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z2<z1)
		lower = z2;
		
	nx = WRAP2(x-1+shift,max);	
	ny = WRAP2(y+1,max) ;
	ii =VECTORIZE(nx,ny,max) ;
	z3 = *(hf+ii) ;
	if (z3<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z3<lower)
		lower = z3;
		
	nx = WRAP2(x-1,max);
	ii = VECTORIZE(nx,y,max)  ;
	z4 = *(hf+ii);
	if (z4<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z4<lower)
		lower = z4;
	
	nx = WRAP2(x-1+shift,max) ;	
	ny = WRAP2(y-1,max) ;
	ii =VECTORIZE(nx,ny,max) ;
	z5 =  *(hf+ii) ;
	if (z5<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z5<lower)
		lower = z5;
		
	nx = WRAP2(x+shift,max);
	ny = WRAP2(y-1,max);
	ii = VECTORIZE(nx, ny,max) ;
	z6 = *(hf+ii);
	if (z6<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z6<lower)
		lower = z6;

// printf("(X,Y): (%d,%d); ninf: %d; lower: %u; threshold: %d; value: %d; ",x,y,ninf, lower, threshold, value);
	if ( (lower+threshold) < value) {
		// Find all the pixels not farther away than "threshold" than the lower one
		for (i=0; i<ninf; i++) { 
			if ((*(hf+v[i])-threshold)<=lower) {
				v[ok] = v[i];
				ok ++;
			}
		}
// printf("ok: %d; ",ok);
		ii = v[rand()%ok];
// printf("ii: %d; ",ii);
		*newx = ii%max;
		*newy = (gint) (ii / max);
// printf("(NEWX,NEWY) = (%d,%d) = (%d,%d)\n", ii%max,(gint) ii/max,*newx,*newy);
		return TRUE;
	}
	else {
// printf("Returning FALSE\n");
		return FALSE;
	}
}

void flow(gint x, gint y, gint tox, gint toy, hf_type *hf, gint max, gboolean wrap) {
	//	Flow a quarter of the H difference between (x,y) and (tox,toy)
	//	from (x,y) to (tox,toy)
//	Define a 3x3 matrix - proportion of the flowing chunk to displace
	static gint matrix[3][3]={1,2,1,2,4,2,1,2,1};
	gint i, toi, m, n, qty, xi, yi;
	hf_type nchunk;

/*	1st version:  simpler... we displace only one pixel... gives some kind of "sand"
	nchunk = *(hf+VECTORIZE(x,y,max)) - *(hf+VECTORIZE(tox,toy,max));
	nchunk = nchunk >> 4;
	i = VECTORIZE(x,y,max);
	toi = VECTORIZE(tox,toy,max);
	*(hf+i) = *(hf+i) - nchunk;
	*(hf+toi) = *(hf+toi) + nchunk;
*/
	nchunk = *(hf+VECTORIZE(x,y,max)) - *(hf+VECTORIZE(tox,toy,max));
	nchunk = nchunk >> 2;
//	nchunk = nchunk >> 4;
	// Remove soil at (x,y)
	for (m=0; m<3; m++)
		for (n=0; n<3; n++) {
			if (wrap)
				i = VECTORIZE(WRAP2(x-1+n,max),WRAP2(y-1+m,max),max);
			else {
				xi = x-1+n;
				yi = y-1+m;
				if ( (xi<0) || (yi<0) || (xi>=max) || (yi>=max) )
					continue;
				else
					i = VECTORIZE(xi,yi,max);
			}
			qty = nchunk * matrix[m][n];
			*(hf+i) = (hf_type) MIN(0xFFFF,MAX(0,(gint) *(hf+i) - qty));
		}
	// Add soil at (tox,toy)
	for (m=0; m<3; m++)
		for (n=0; n<3; n++) {
			if (wrap)
				i = VECTORIZE(WRAP2(tox-1+n,max),WRAP2(toy-1+m,max),max);
			else {
				xi = tox-1+n;
				yi = toy-1+m;
				if ( (xi<0) || (yi<0) || (xi>=max) || (yi>=max) )
					continue;
				else
					i = VECTORIZE(xi,yi,max);
			}
			qty = nchunk * matrix[m][n];
			*(hf+i) = (hf_type) MIN(0xFFFF,MAX(0,(gint) *(hf+i) + qty));
		}

}

void hf_rain_erosion_hex (hf_struct_type *hf, gint drops, gint slope_threshold) {
//	Simple iterative erosion
//	using hexagonal sampling for a more natural looking result
//	We generate random drops and follow their path
//	We don't flow a drop when the slope is more than 2 power thresh
//	Neighbours are numbered from 1 to 6, clockwise, 1 being (1,0)
//	Not extensively tested when the HF doesn't wrap
	
	static gint x,y, i, ii, s, val, threshold, nextx, nexty, lngpath;
	unsigned long int steps;
	hf_type *tmp;
	gboolean even=TRUE;
	
//	threshold is the future chunk... it must be at least 16, because the center
//	of the 3x3 averaging matrix == 4 and we divide the chunk by 16 before applying it
//	threshold = 15 + (gint) pow(2.0,(double) thresh);
	threshold = MAX(16,iget_absolute_slope(slope_threshold,hf->max_x));
// printf("RAIN: Threshold (slope): %d = %d\n",slope_threshold,threshold);
//	In the 1st version, drops were given as an exponent
//	We could fallback to this in the future...
//	drops = (long int) pow(2.0,(gdouble) dr_log);
	if (!hf->tmp_buf)
		hf_backup(hf);
		
	tmp = hf->hf_buf;
	hf->hf_buf = hexagonal_row_sampling (hf->hf_buf, hf->max_x, 
		hf->if_tiles, even, TRUE); 
	free(tmp);
	
// printf("RAIN EROSION, DROPS: %d; THRESHOLD: %d\n", drops, threshold);
	for (s=0; s<drops; s++) {
		x = rand()%hf->max_x;
		y = rand()%hf->max_y;
		// Make the (x,y) drop flow until it encounters a minimum
		steps = 0;
// printf("********** DROP: %u; (x,y): (%d,%d)\n",s,x,y);
		while (find_all_neighbours_hex(x,y,&nextx,&nexty,
				hf->max_x,hf->hf_buf,threshold, TRUE, TRUE))	{
			if (!hf->if_tiles)
				if ( (x==0) || (y==0) ||
					(x==(hf->max_x-1)) ||
					(y==(hf->max_y-1)) )
						break;
			flow(x,y,nextx,nexty,hf->hf_buf,hf->max_x, hf->if_tiles);
			ii = VECTORIZE(nextx,nexty,hf->max_x);
			y = nexty;
			x = nextx;
			steps++;
			if (steps>hf->max_x)
				break;
		}
	}		
	// Restore the rectangular sampled hf
	tmp = hf->hf_buf;
	hf->hf_buf = hexagonal_row_sampling (hf->hf_buf, hf->max_x, 
		hf->if_tiles, even, FALSE); 
	free(tmp);
}


void hf_crests_erosion_hex (hf_struct_type *hf, gint steps, gint slope_threshold) {
//	Simple iterative erosion, giving crested mountains
//	Hexagonally sampled version (see hf_rain_erosion_hex) 2005-02
//	We find the minimum neighbour
//	If its value is superior to the current pixel one,
//	we add half the difference to the current pixel
//	If the minimum neighbour is inferior, we subtract half the current pixel
//	This is like if half the difference was "flowing" from or to the current pixel
//	Neighbours are numbered from 1 to 6, clockwise, 1 being (1,0)
//	2004-03-11 Corrected to deal with wrapping

	static gint z1, z2, z3, z4, z5, z6;
	static gint x,y, i, s, val, ii, minii, lower, threshold, shift;
	hf_type value, *tmp;
	gboolean test, even=TRUE;
	gfloat ratio; // Factor for dividing the weight of the minimum pixel, when it is on the diagonals
	
	// Slope_threshold is in degrees
	threshold = iget_absolute_slope(slope_threshold,hf->max_x);
// printf("Threshold (slope): %d = %d\n",slope_threshold,threshold);
	if (!hf->tmp_buf)
		hf_backup(hf);
			
	tmp = hf->hf_buf;
	hf->hf_buf = hexagonal_row_sampling (hf->hf_buf, hf->max_x, 
		hf->if_tiles, even, TRUE); 
	free(tmp);
		
	test = y & (gint) 1; // test = TRUE if value is odd
	if ( (test && even) || ((!test) && (!even)) )  //  XOR
		shift = 1;
	else
		shift = 0;
	
	// We need a second temporary buffer for storing
	// the result of each step
	if (!hf->result_buf) {
		hf->result_buf = (hf_type *)malloc(hf->max_x*hf->max_y*sizeof(hf_type));
	}
// printf("CRESTS EROSION, STEPS: %d; THRESHOLD: %d\n", steps, threshold);
	memcpy(hf->result_buf, hf->tmp_buf, hf->max_x*hf->max_y*sizeof(hf_type));
	for (s=0; s<steps; s++) {
	for (y=0; y<hf->max_y; y++) {
		for (x=0; x<hf->max_x; x++) {
			i = VECTORIZE(x,y,hf->max_x);
			value = *(hf->hf_buf + i);
			if (hf->if_tiles)
				ii=VECTORIZE(WRAP2(x+1,hf->max_x),y,hf->max_x);
			else
				ii=VECTORIZE(MIN(x+1,hf->max_x-1),y,hf->max_x);
			z1 =  *(hf->hf_buf+ii);
			minii = ii;
			lower = z1;
			ratio = 1.0;
			
			if (hf->if_tiles)
				ii=VECTORIZE(WRAP2(x+shift,hf->max_x), WRAP2(y+1,hf->max_y),hf->max_x) ;
			else
				ii=VECTORIZE(MIN(x+shift,hf->max_x-1), MIN(y+1,hf->max_y-1),hf->max_x) ;
			z2 = *(hf->hf_buf+ii);
			if (z2<z1) {
				minii = ii;
				lower = z2;
				ratio = 1.155;
			}
			
			if (hf->if_tiles)
				ii =VECTORIZE( WRAP2(x-1+shift,hf->max_x), WRAP2(y+1,hf->max_y),hf->max_x) ;
			else
				ii = VECTORIZE(MAX(0,x-1+shift),MIN(y+1,hf->max_y-1),hf->max_x) ;
			z3 = *(hf->hf_buf+ii) ;
			if (z3<lower) {
				minii = ii;
				lower = z3;
				ratio = 1.155;
			}
			
			if (hf->if_tiles)
				ii=VECTORIZE(WRAP2(x-1,hf->max_x), y, hf->max_x)  ;
			else
				ii=VECTORIZE(MAX(x-1,0), y, hf->max_x)  ;
			z4 = *(hf->hf_buf+ii);
			if (z4<lower) {
				minii = ii;
				lower = z4;
				ratio = 1.0;
			}
			
			if (hf->if_tiles)
				ii=VECTORIZE(WRAP2(x-1+shift,hf->max_x),
				 WRAP2(y-1,hf->max_x),hf->max_x) ;
			else
				ii=VECTORIZE(MAX(x-1+shift,0), MAX(y-1,0) ,hf->max_x) ;
			z5 =  *(hf->hf_buf+ii) ;
			if (z5<lower) {
				minii = ii;
				lower = z5;
				ratio = 1.155;
			}
			
			if (hf->if_tiles)
				ii=VECTORIZE(WRAP2(x+shift,hf->max_x),WRAP2(y-1,hf->max_y),hf->max_x) ;
			else
				ii=VECTORIZE(MIN(x+1,hf->max_x-1),MAX(y-1,0),hf->max_x) ;
			z6 = *(hf->hf_buf+ii);
			if (z6<lower) {
				minii = ii;
				lower = z6;
				ratio = 1.155;
			}
			
			val = ( ( ((gint) *(hf->hf_buf+minii)) - (gint) value ) / 2);
			if (threshold < ABS(val)) {
				*(hf->result_buf+i) = value + (gint) ( ((gfloat) val) / ratio);
//				*(hf->hf_buf+minii) -= (gint) ( ((gfloat) val) / ratio);
			}
			else
				*(hf->result_buf+i) = value;
		}
	}
	memcpy(hf->hf_buf, hf->result_buf, hf->max_x*hf->max_y*sizeof(hf_type));
	}
	
	tmp = hf->hf_buf;
	hf->hf_buf = hexagonal_row_sampling (hf->hf_buf, hf->max_x, 
		hf->if_tiles, even, FALSE); 
	free(tmp);
// printf("MAX: %d; MIN: %d; DELTA: %d\n",hf->max, hf->min, hf->max-hf->min);

}

gboolean dfind_all_neighbours( gint x, gint y, gint *newx, gint *newy,
	gint max, gdouble *hf, gdouble threshold) {
	//	Find all the inferior neighbours of (x,y) in hf
	//	Randomly select one of the neighbours,
	//	given its height difference with the minimum is < to threshold
	//	Returns the result in newx and newy
	//	If (x,y) is a minimum, returns FALSE
	static gint v[8], i, ii, nx, ny, ninf, ok;
	gdouble value,z1, z2, z3, z4, z5, z6, z7, z8;
	gdouble lower= (gdouble) MAX_HF_VALUE;
	ninf = 0;
	ok = 0;
	value = *(hf+VECTORIZE(x,y,max)) ;
	ii=VECTORIZE(WRAP2(x+1,max),y,max);
	z1 =  *(hf+ii);
	if (z1<value) {
		v[ninf] = ii;
		ninf++;
	}
	lower = z1;
	nx = WRAP2(x+1,max);
	ny = WRAP2(y+1,max);
	ii=VECTORIZE(nx, ny,max) ;
	z2 = *(hf+ii);
	if (z2<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z2<z1)
		lower = z2;
	ny=WRAP2(y+1,max) ;
	ii =VECTORIZE(x,ny,max) ;
	z3 = *(hf+ii) ;
	if (z3<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z3<lower)
		lower = z3;
	nx = WRAP2(x-1,max);
	ny = WRAP2(y+1,max);
	ii=VECTORIZE(nx,ny,max)  ;
	z4 = *(hf+ii);
	if (z4<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z4<lower)
		lower = z4;
	nx = WRAP2(x-1,max) ;
	ii =VECTORIZE(nx,y,max) ;
	z5 =  *(hf+ii) ;
	if (z5<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z5<lower)
		lower = z5;
	nx = WRAP2(x-1,max);
	ny = WRAP2(y-1,max);
	ii=VECTORIZE(nx, ny,max) ;
	z6 = *(hf+ii);
	if (z6<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z6<lower)
		lower = z6;
	ny = WRAP2(y-1,max) ;
	ii =VECTORIZE(x,ny,max) ;
	z7 = *(hf+ii);
	if (z7<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z7<lower)
		lower = z7;
	nx = WRAP2(x+1,max);
	ny = WRAP2(y-1,max) ;
	ii=VECTORIZE(nx,ny,max)  ;
	z8 = *(hf+ii);
	if (z8<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z8<lower)
		lower = z8;

// printf("(%d, %d) => LOWER: %5.2f; THRESHOLD: %5.2f; VALUE: %5.2f\n", x, y, lower, threshold, value);
	if ( (lower+threshold) < value) {
		// Find all the pixels not farther away than "threshold" than the lower one
		for (i=0; i<ninf; i++) { 
			if ((*(hf+v[i])-threshold)<=lower) {
				v[ok] = v[i];
				ok ++;
			}
		}
// printf("ok: %d; ",ok);
		ii = v[rand()%ok];
// printf("ii: %d; ",ii);
		*newx = ii%max;
		*newy = (gint) (ii / max);
// printf("(NEWX,NEWY) = (%d,%d) = (%d,%d)\n", ii%max,(gint) ii/max,*newx,*newy);
		return TRUE;
	}
	else {
// printf("Returning FALSE\n");
		return FALSE;
	}
}
void dadd_rain (gdouble *water, gint max_x, gint max_y, gdouble *v, gint vlength) {
	gint x,y;
	for (y=0; y<max_y; y++) {
		for (x=0; x<max_x; x++) {
			// Generate water  
			*(water+VECTORIZE(x,y,max_x)) += v[rand()%vlength];
		}
	}
}

static gdouble *build_map (gint soil_qty, gint radius) {

	gdouble *map, sum, dval, ratio;
	gint i,j,map_size;

	map_size = (2*radius) + 1;
	map = (gdouble *) malloc(map_size*map_size*sizeof(gdouble));
	sum = 0.0;
	if (!radius)
		sum = (gdouble) soil_qty;
	else {
		for (i=-radius; i<radius+1; i++) {
			for (j=-radius; j<radius+1; j++) {
				dval = BELLD(CONST_E,2.0,i,radius) * 
					BELLD(CONST_E,2.0,j,radius);
				sum += dval;
			}
		}
	}

	ratio = ((gdouble) soil_qty) / sum;
/*
	printf("SUM: %6.3f; RADIUS: %5d; ",sum, radius);
	printf("SOIL_QTY: %5d ",soil_qty);
	printf("RATIO: %6.3f\n", ratio);
*/	
//	printf("MAP %d x %d:\n", (2*radius) + 1, (2*radius) + 1);
	for (i=-radius; i<radius+1; i++) {
		for (j=-radius; j<radius+1; j++) {
			*(map+VECTORIZE(i+radius,j+radius,map_size)) = ratio * BELLD(CONST_E,2.0,i,radius) * 		BELLD(CONST_E,2.0,j,radius);
//			if (radius<4) printf("(%d, %d) == %5.2f;  ",i,j,*(map+VECTORIZE(i,j,map_size)) );
		}
//		if (radius<4) printf("\n");
	}
	return map;
}

void dflow_map(gint x, gint y, gint tox, gint toy, gdouble *hf, gdouble *water, gint max, gboolean wrap) {
	// Flow a quarter of the water (part considered as soil in suspension)
	// between (x,y) and (tox,toy) from (x,y) to (tox,toy)

	gint i, fromi, toi, isoil, radius, map_size;
	
	gdouble soil, *map;
		
	// 1. Check if required map is already cached
	// 2. Build it if not, add to the cache
	
	fromi = VECTORIZE(x,y,max);
	toi = VECTORIZE(tox,toy,max);
	
	soil = *(water + fromi);
	
// printf("FROM (%d, %d) TO (%d, %d); SOIL: %5.2f; \n", x, y, tox, toy, soil);
	isoil = MIN(MAX_CACHE-1,(gint) soil);

	if (!isoil)
		return;
	
	radius = (gint) log((gdouble) isoil);
	map_size = 2*radius +1;
/*	
	map = *(cache+isoil);
	
	if (! map ) {
		map = build_map (isoil, radius);
		*(cache+isoil) = map;
	}
*/
//	if (map_size==1)
		*(water+VECTORIZE(tox,toy,max)) += soil;
//	else
//		generalized_merge ((gpointer) map, GDOUBLE_ID, map_size, map_size, (gpointer) water, GDOUBLE_ID, max, max, tox, toy, ADD, wrap, FALSE);
/*

	if (map_size==1) {
		*(hf+VECTORIZE(x,y,max)) -= soil;
		*(hf+VECTORIZE(tox,toy,max)) += soil;
	}	
	else {
		generalized_merge ((gpointer) map, GDOUBLE_ID, map_size, map_size, (gpointer) hf, GDOUBLE_ID, max, max, x, y, SUBTRACT, wrap, FALSE);
	
		generalized_merge ((gpointer) map, GDOUBLE_ID, map_size, map_size, (gpointer) hf, GDOUBLE_ID, max, max, tox, toy, ADD, wrap, FALSE);
	}


	// We flow all the water mass
	*(water+toi) += *(water+fromi);
	*(water+fromi) = 0.0;
*/
}

void check_water (gdouble *water, gint max) {

	gint i, count=0;
	gdouble sum=0.0;
	
	for (i=0; i<max*max; i++) {
		if (*(water+i)>0.0) {
			count ++;
			sum += *(water+i);
		}
	}
	printf("WATER PIX > 0: %d; SUM: %5.2f; AVRG: %5.2f\n",count, sum, sum / (gdouble) count);

}

void check_hf (gdouble *hf, gint max) {

	gint i, countneg=0, countoverflow=0;
	gdouble sumo=0.0, sumn=0.0;
	
	for (i=0; i<max*max; i++) {
		if (*(hf+i)<1.0) {
			countneg ++;
			sumn += *(hf+i);
		}
		else
		if (*(hf+i)>(gdouble)MAX_HF_VALUE) {
			countoverflow ++;
			sumo += *(hf+i) - (gdouble) MAX_HF_VALUE;
		}
		
		
	}
	printf("HF < 1: %d; AVRG: %7.2f HF > MAX: %d; AVRG: %7.2f\n",countneg, sumn / (gdouble) countneg, countoverflow, sumo / (gdouble) countoverflow);

}

void hf_water_erosion (hf_struct_type *hf, gint steps, gint slope_threshold) {
	// Water / fluvial erosion
	// Simple model without speed and direction matrix
	// Derivated from the rain erosion
	// Basically, we add drops in a water matrix
	// We never spread water explicitly
	// Howvever, we spread soil with a gaussian matrix which size depends
	// on the water weight
	// Buffers:
	//	input, output: height field converted to gdouble, 0 - 0xFFFF -> 0 - 1.0
	//	water = water mass;
	gdouble threshold, w, diff;
	gdouble *water,*input, *output;
	gint i,s,x,y, nextx, nexty;
	gboolean even=TRUE;
	gdouble vpow2[15] = {1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 4.0, 4.0 ,8.0};
	
//	for (i=0; i<15; i++)
//		vpow2[i] /= (gdouble) MAX_HF_VALUE;
		
	if (!if_cache_init) {
//		printf("@@@@@@@@@@@@@@@@ Initializing CACHE @@@@@@@@@@@@@@@@\n");
		cache = (gdouble **) malloc(MAX_CACHE*sizeof(gdouble *));
		for (i=0; i<MAX_CACHE; i++) {
			*(cache+i) = NULL;
		}
		if_cache_init = TRUE;
	}
	
	threshold = dget_absolute_slope(slope_threshold,hf->max_x);
	
	input = alloc_double_hf(hf->max_x);
	water = calloc_double_hf(hf->max_x);
	
	hf_type_to_double (hf->hf_buf, hf->max_x, hf->max_y, input);
	
	// For each step, for each pixel:
	//	1. Add rain drops to the current water mass
	//		Choice at random
	//		water mass = power of 2, from 0 to 3), small chunks more frequent
	//	2. Flow the water mass with a simple rule, like we do with rain
	//		Modify the water matrix
	//		Add / subtract the soil from input to output
	
	for (s=0; s<steps; s++) {
		printf("STEP : %d / %d\n",s, steps);
		dadd_rain (water, hf->max_x, hf->max_y, vpow2, 15);
		
		for (y=0; y<hf->max_y; y++) {
			for (x=0; x<hf->max_x; x++) {
				i = VECTORIZE(x,y,hf->max_x);

				if (dfind_all_neighbours(x,y,&nextx,&nexty,
						hf->max_x,input,threshold))	{
					if (!hf->if_tiles)
						if ( (x==0) || (y==0) ||
							(x==(hf->max_x-1)) ||
							(y==(hf->max_y-1)) )
								break;
//					printf("Flowing from (%d, %d) to (%d, %d)\n",x,y,nextx,nexty);
					dflow_map(x,y,nextx,nexty,input,water,hf->max_x, hf->if_tiles);
				}
			}
		}
		for (i=0; i<hf->max_x*hf->max_y; i++)
			*(input+i) = *(water+i) + *(input+i);
//		check_water (water, hf->max_x);
//		check_hf (input, hf->max_x);
	}
	
	double_clamp (hf->hf_buf, hf->max_x, hf->max_y, 0, MAX_HF_VALUE, water);
//	double_clamp (hf->hf_buf, hf->max_x, hf->max_y, 0, MAX_HF_VALUE, input);

	free(input);
	free(water);
}

/*
void hf_water_erosion_hex (hf_struct_type *hf, gint steps, gint slope_threshold) {
	// Water / fluvial erosion
	// Simple model without speed and direction matrix
	// Derivated from the rain erosion
	// Basically, we add drops in a water matrix
	// We never spread water explicitly
	// Howvever, we spread soil with a gaussian matrix which size depends
	// on the water weight
	// Buffers:
	//	input, output: height field converted to gdouble, 0 - 0xFFFF -> 0 - 1.0
	//	water = water mass;
	gdouble threshold, w, diff;
	hf_type *water,*input, *output;
	hf_type *tmp;
	gint i,s,x,y, nextx, nexty;
	gboolean even=TRUE;
	gdouble vpow2[15] = {1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 4.0, 4.0 ,8.0};
	
	for (i=0; i<15; i++)
		vpow2[i] /= (gdouble) MAX_HF_VALUE;
		
	threshold = dget_absolute_slope(slope_threshold,hf->max_x);			

	input = hexagonal_row_sampling (hf->hf_buf, hf->max_x, 
		hf->if_tiles, even, TRUE);
	
	water = calloc_hf_type(hf->max_x);

	// For each step, for each pixel:
	//	1. Add rain drops to the current water mass
	//		Choice at random
	//		water mass = power of 2, from 0 to 3), small chunks more frequent
	//	2. Flow the water mass with a simple rule, like we do with rain
	//		Modify the water matrix
	//		Add / subtract the soil from input to output
	
	for (s=0; s<steps; s++) {
		add_rain (water, max_x, max_y, vpow2, 15);
		
		for (y=0; y<hf->max_y; y++) {
			for (x=0; x<hf->max_x; x++) {
				i = VECTORIZE(x,y,hf->max_x);

				if (find_all_neighbours_hex(x,y,&nextx,&nexty,
						hf->max_x,input,threshold, TRUE, TRUE))	{
					if (!hf->if_tiles)
						if ( (x==0) || (y==0) ||
							(x==(hf->max_x-1)) ||
							(y==(hf->max_y-1)) )
								break;
					flow_map(x,y,nextx,nexty,input,water,hf->max_x, hf->if_tiles);
				}
			}
		}
	}
	
	tmp = hexagonal_row_sampling (input, hf->max_x, 
		hf->if_tiles, even, FALSE); 
	
	memcpy(hf->hf_buf, tmp, hf->max_x*hf->max_y*sizeof(hf_type));
	
	free(tmp);
	free(input);
	free(water);
}
*/
/*
void flow_water_hex (gdouble *hf, gdouble *water, gdouble *direction, gint x, gint y, gint wrap, gboolean even, gboolean shift_right) {

//	Flowing water from x,y to lower surrounding pixels
//	We add water and ground heights to get the total height

	static gint v0[6] = {0,0,0,0,0,0};
	static gint i, x,y,v[6],z1, z2, z3, z4, z5, z6, i, ii, nx, ny, ninf, shift, sumdif;
	gdouble value;
	gboolean test;

	ninf = 0;
	sumdif = 0;
	memcpy(v,v0,6*sizeof(gint));
	i = VECTORIZE(x,y,max_x);
	value = *(hf+i) + *(water+i);
	test = y & (gint) 1; // test = TRUE if value is odd
	if ( (test && even) || ((!test) && (!even)) )  //  XOR
		shift = 1;
	else
		shift = 0;

	ii=VECTORIZE(WRAP2(x+1,max_x),y,max_x);
	z1 =  *(hf+ii);
	if (z1<value) {
		if (wrap || ( (x+1)<max_x ) ) {
			v[ninf] = ii;
			ninf++;
			sumdif += value-z1;
		}
	}
	
	nx = WRAP2(x+shift,max_x);
	ny = WRAP2(y+1,max_y);
	ii=VECTORIZE(nx, ny,max_x) ;
	z2 = *(hf+ii);
	if (z2<value) {
		if (wrap || (((x+shift)<max_x) && ((y+1)<max_y)) ) {
			v[ninf] = ii;
			ninf++;
			sumdif += value-z2;
		}
	}
		
	nx = WRAP2(x-1+shift,max_x);	
	ny = WRAP2(y+1,max_y) ;
	ii =VECTORIZE(nx,ny,max_x) ;
	z3 = *(hf+ii) ;
	if (z3<value) {
		if (wrap || (((x+shift)>0) && ((y+1)<max_y)) ) {
			v[ninf] = ii;
			ninf++;
			sumdif += value-z3;
		}
	}
		
	nx = WRAP2(x-1,max_x);
	ii = VECTORIZE(nx,y,max_x)  ;
	z4 = *(hf+ii);
	if (z4<value) {
		if (wrap || (x>0)) {
			v[ninf] = ii;
			ninf++;
			sumdif += value-z4;
		}
	}
	
	nx = WRAP2(x-1+shift,max_x) ;	
	ny = WRAP2(y-1,max_y) ;
	ii = VECTORIZE(nx,ny,max_x) ;
	z5 =  *(hf+ii) ;
	if (z5<value) {
		if (wrap || ((y>0) && ((x+shift)>0)) ) {
			v[ninf] = ii;
			ninf++;
			sumdif += value-z5;
		}
	}
		
	nx = WRAP2(x+shift,max_x);
	ny = WRAP2(y-1,max_y);
	ii = VECTORIZE(nx, ny,max_x) ;
	z6 = *(hf+ii);
	if (z6<value) {
		if (wrap || ((y>0) && ((x+shift)<max_x)) ) { 
			v[ninf] = ii;
			ninf++;
			sumdif += value-z6;
		}
	}

	if (!ninf)
		return;			
	
	*(hf+VECTORIZE(x,y,max_x)) -= sumdif / (ninf+1);
	
	sumdif = sumdif / (ninf*(ninf+1));
	
	for (i=0; i<ninf; i++) {
		*(hf+v[i]) += sumdif;
	}

}

void hf_water_erosion_hex1 (hf_struct_type *hf, gint steps, gint slope_threshold) {
	// Water / fluvial erosion
	// Buffers:
	//	input, output: height field converted to gdouble, 0 - 0xFFFF -> 0 - 1.0
	//	water = water mass;
	//	susp = suspended soil (<= water mass);
	//	speed = flow speed;
	//	direction = flow direction (radians);
	gdouble threshold, w, diff, satur_max, SATUR_K = 1.0;
	gdouble *water, *susp, *speed, *direction, *input, *output;
	hf_type *tmp;
	gint i,s,x,y;
	gboolean even=TRUE;
	gdouble vpow2[15] = {1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 4.0, 4.0 ,8.0};
	
	for (i=0; i<15; i++)
		vpow2[i] /= (gdouble) MAX_HF_VALUE;
		
	threshold = dget_absolute_slope(slope_threshold,hf->max_x);			
		
	tmp = hf->hf_buf;
	hf->hf_buf = hexagonal_row_sampling (hf->hf_buf, hf->max_x, 
		hf->if_tiles, even, TRUE); 
	free(tmp);
	
	input = calloc_double_hf(hf->max_x);
	output = calloc_double_hf(hf->max_x);
	water = calloc_double_hf(hf->max_x);
	speed = calloc_double_hf(hf->max_x);
	direction = calloc_double_hf(hf->max_x);
	susp = calloc_double_hf(hf->max_x);
	
	hf_type_to_double (hf->hf_buf, hf->max_x, hf->max_y, input);
	
	// For each step, for each pixel:
	//	1. Add water or not
	//		Choice at random
	//		water mass = power of 2, from 0 to 3), small chunks more frequent
	//	2. Add to the current water mass
	//	3. Check where we flow the water mass
	//		Balance the water mass between the pixels (current + neighbours) 
	//		Calculate the new saturation level, given the speed,
	//		for all the pixels
	//		Deposit the excess of soil, or remove the deficit
	//		We spread only the water mass
	//		The neighbours are not tested for deposit,
	//		because they would be tested as "current" later
	
	for (s=0; s<steps; s++) {
		for (y=0; y<hf->max_y; y++) {
			for (x=0; x<hf->max_x; x++) {
				i = VECTORIZE(x,y,hf->max_x);
				// Generate water
				w =  vpow2[rand()%15];
				*(water+i) += w;
				// This water has a basic speed, proportional to its volume
				// This initial speed has no effect on the direction
				// because this water comes perpendicularly to the ground,
				// from the sky or from the ground
				// We fix arbitrarily the basic speed to 1.0, when
				// we use the speed for calculating the saturation level
				// Now for checking if we need to remove or deposit soil,
				// we calculate the saturation level, given
				// speed and water mass, then remove soil up to it
				// (transfer it to susp)
				satur_max = *(water+i) * (1.0 + *(speed+i)) * SATUR_K;
				diff = *(susp+i) - satur_max;
				*(susp+i) = satur_max;
				*(output+i) = *(input+i) + diff; 
				// Spread water and soil
				// Spread the speed vector, accelerated by the slope
				
			}
		}
		// We swap input and output for the next step
		swap_buffers ((gpointer *) &input, (gpointer *) &output);
	}
	
	//	More recent results are always in "input"
	double_clamp (hf->hf_buf, hf->max_x, hf->max_y, 0, MAX_HF_VALUE, input);

	tmp = hf->hf_buf;
	hf->hf_buf = hexagonal_row_sampling (hf->hf_buf, hf->max_x, 
		hf->if_tiles, even, FALSE); 
		
	free(tmp);

	free(water);
	free(susp);
	free(speed);
	free(direction);
	free(input);
	free(output);

}
*/

/*
void hf_slope (hf_struct_type *hf) {
//	Calculate the average slope (tangent) of each point
//	Output the length of the tangent vector in hf_buf and display it
//	Purpose:  to explore the result for eventually building eroded HFs...
//	This one gives thin lines at the crests top or bottom, where the sum
//	of slopes calculated from different xy angles is null
//	The HF should be smoothed before to see the pattern (radius ~ 3 to 10)
//	Lines are black for the top (maxima) and white for the bottom (minima)
//	We simply average all the vectors given by
//	the difference between a pixel and its 8 neighbours
//	Neighbours are numbered from v1 to v8, clockwise, v1 being (1,0)
//	The default for this version of the program is to wrap
	static gfloat l1, l2, l3, l4, l5, l6, l7, l8;
	static gfloat dz1, dz2, dz3, dz4, dz5, dz6, dz7, dz8;
	hf_type value;
	static gint x,y, i;
	static vector v;
	static gfloat min, ratio, val;
	static gfloat maxf = (gfloat) 0xFFFF;
	static gfloat revmaxf;
	static gfloat max = 0.0;
	gfloat *result;
	result = (gfloat *) malloc(sizeof(gfloat) * hf->max_x *hf->max_y);
	min = maxf;
	revmaxf = 0.001;
	if (!hf->tmp_buf)
		hf_backup(hf);
	for (y=0; y<hf->max_y; y++)
		for (x=0; x<hf->max_x; x++) {
			i = VECTORIZE(x,y,hf->max_x);
			value = *(hf->tmp_buf + i);

			dz1 = (gfloat) *(hf->tmp_buf+VECTORIZE(WRAP2(x+1,hf->max_x),y,
				hf->max_x)) - value;
			dz2 = (gfloat) *(hf->tmp_buf+VECTORIZE(WRAP2(x+1,hf->max_x),
				WRAP2(y+1,hf->max_y),hf->max_x)) - value;
			dz3 = (gfloat) *(hf->tmp_buf+VECTORIZE(x,WRAP2(y+1,hf->max_y),
				hf->max_x)) - value;
			dz4 = (gfloat) *(hf->tmp_buf+VECTORIZE(WRAP2(x-1,hf->max_x),
				WRAP2(y+1,hf->max_y),hf->max_x)) - value;
			dz5 = (gfloat) *(hf->tmp_buf+VECTORIZE(WRAP2(x-1,hf->max_x),
				y,hf->max_x)) - value;
			dz6 = (gfloat) *(hf->tmp_buf+VECTORIZE(WRAP2(x-1,hf->max_x),
				WRAP2(y-1,hf->max_y),hf->max_x)) - value;
			dz7 = (gfloat) *(hf->tmp_buf+VECTORIZE(x,WRAP2(y-1,hf->max_y),
				hf->max_x)) - value;
			dz8 = (gfloat) *(hf->tmp_buf+VECTORIZE(WRAP2(x+1,hf->max_x),
				WRAP2(y-1,hf->max_y),hf->max_x)) - value;
			l1 = 1.0 / sqrt(1.0+dz1*dz1);
			l2 = 1.0 / sqrt(2.0+dz2*dz2);
			l3 = 1.0 / sqrt(1.0+dz3*dz3);
			l4 = 1.0 / sqrt(2.0+dz4*dz4);
			l5 = 1.0 / sqrt(1.0+dz5*dz5);
			l6 = 1.0 / sqrt(2.0+dz6*dz6);
			l7 = 1.0 / sqrt(1.0+dz7*dz7);
			l8 = 1.0 / sqrt(2.0+dz8*dz8);

			v.x = l1 + l2 - l4 - l5 - l6 + l8;  // x == 0 for v3 and v7;
			v.y = l2 + l3 + l4 - l6 - l7 - l8;  // y == 0 for v1 and v5;
			v.z = dz1*l1 + dz2*l2 + dz3*l3 + dz4*l4 + dz5*l5 + dz6*l6 + dz7*l7 + dz8*l8;
// printf("(%d,%d): v.z: %8.5f; dz1: %8.1f, dz2: %8.1f, dz3: %8.1f, dz4: %8.1f, dz5: %8.1f, dz6: %8.1f, dz7: %8.1f, dz8: %8.1f \n", 
//  x,y,v.z, dz1,dz2,dz3,dz4,dz5,dz6,dz7,dz8);
// printf("(%d,%d): %f,%f,%f; l1: %f; l2: %f; l3: %f; l4: %f; l5: %f; l6: %f; l7: %f; l8: %f;\n",x,y,v.x, v.y, v.z,l1, l2, l3, l4, l5, l6, l7, l8);
			*(result+ i) = v.z;
			if (v.z>max)
				max = v.z;
			else
				if (v.z<min)
					min = v.z;
		}
		ratio = maxf / (max - min);
//		printf("MAX: %8.5f; MIN: %8.5f; DELTA: %8.5f\n",max, min, max-min);
		for (i=0; i< hf->max_x*hf->max_y; i++) {
			val = ratio * (*(result+i) - min);
			*(hf->hf_buf+i) = (hf_type) val;
		}
	free(result);
}
*/
/*
void hf_crests (hf_struct_type *hf) {
//	Builds a "cresting" template by replacing each pixel with its
//	neighbour having the minimum difference with it
//	Crests also appear where the curvature is negative... (bottoms)
//	Neighbours are numbered from 1 to 8, clockwise, 1 being (1,0)
//	The default for this version of the program is to wrap
//	Kept here for documentation purpose
//	hf_crests_erosion is more realistic
	static gint z1, z2, z3, z4, z5, z6, z7, z8;
	static gint x,y, i, val;
	hf_type value;
	if (!hf->tmp_buf)
		hf_backup(hf);
	for (y=0; y<hf->max_y; y++)
		for (x=0; x<hf->max_x; x++) {
			i = VECTORIZE(x,y,hf->max_x);
			value = *(hf->tmp_buf + i);
			z1 =  *(hf->tmp_buf+VECTORIZE(WRAP2(x+1,hf->max_x),y,hf->max_x))  ;
			z2 = *(hf->tmp_buf+VECTORIZE(WRAP2(x+1,hf->max_x),
				WRAP2(y+1,hf->max_y),hf->max_x)) ;
			z3 = *(hf->tmp_buf+VECTORIZE(x,WRAP2(y+1,hf->max_y),hf->max_x))  ;
			z4 = *(hf->tmp_buf+VECTORIZE(WRAP2(x-1,hf->max_x),
				WRAP2(y+1,hf->max_y),hf->max_x))  ;
			z5 =  *(hf->tmp_buf+VECTORIZE(WRAP2(x-1,hf->max_x),y,hf->max_x)) ;
			z6 = *(hf->tmp_buf+VECTORIZE(WRAP2(x-1,hf->max_x),
				WRAP2(y-1,hf->max_y),hf->max_x)) ;
			z7 = *(hf->tmp_buf+VECTORIZE(x,WRAP2(y-1,hf->max_y),hf->max_x)) ;
			z8 = *(hf->tmp_buf+VECTORIZE(WRAP2(x+1,hf->max_x),
				WRAP2(y-1,hf->max_y),hf->max_x)) ;
// printf("z1: %d, z2: %d, z3: %d, z4: %d, z5: %d, z6: %d, z7: %d, z8: %d\n",z1, z2, z3, z4, z5, z6, z7, z8);
//			For crests - displaced minimal variation
			val = MIN(z1,MIN(z2,MIN(z3,MIN(z4,MIN(z5,MIN(z6,MIN(z7,z8))))))) - value;
			*(hf->hf_buf+i) = 0x1000 + val;
		}
	hf_auto_contrast(hf);
//	printf("MAX: %d; MIN: %d; DELTA: %d\n",hf->max, hf->min, hf->max-hf->min);
}
*/
/*
void hf_crests_erosion1 (hf_struct_type *hf, gint steps, gint thresh) {
//	GIVES TOP-DOWN SMOOTHING WITH VERY SLIGHT CRESTS
//	One variation gives oriented erosion (see comments at the end)
//	Averaging the neighbours > to current value
//	Neighbours are numbered from 1 to 8, clockwise, 1 being (1,0)
//	The default for this version of the program is to wrap
	static gint z1, z2, z3, z4, z5, z6, z7, z8;
	static long int totsup, value, nsup;
	gint *deltamap;
	static gint v[8], p[8], vi, x,y, i, s, val, ii, minii, lower, threshold, k, l, m;
	deltamap = (gint *) malloc(sizeof(gint) * hf->max_x * hf->max_y);
//	threshold = (0xFFFF / hf->max_x) << 2;
	threshold = (gint) pow(2.0,(double) thresh);
	if (!hf->tmp_buf)
		hf_backup(hf);
printf("CRESTS EROSION, STEPS: %d; THRESHOLD: %d\n", steps, threshold);
	for (s=0; s<steps; s++) {
		for (i=0; i<hf->max_x*hf->max_y; i++)
			*(deltamap+i) = 0;
	for (y=0; y<hf->max_y; y++) {
		for (x=0; x<hf->max_x; x++) {
			nsup = 0;
			totsup = 0;
			i = VECTORIZE(x,y,hf->max_x);
			value = *(hf->hf_buf + i);
			ii=VECTORIZE(WRAP2(x+1,hf->max_x),y,hf->max_x);
			z1 =  *(hf->hf_buf+ii);
			if (z1>(value+threshold)) {
				totsup+=z1;
				nsup++;
				v[nsup-1] = ii;
				p[nsup-1]=z1;
			}
			ii=VECTORIZE(WRAP2(x+1,hf->max_x), WRAP2(y+1,hf->max_y),hf->max_x) ;
			z2 = *(hf->hf_buf+ii);
			if (z2>(value+threshold)) {
				totsup+=z2;
				nsup++;
				v[nsup-1] = ii;
				p[nsup-1]=z2;
			}
			ii =VECTORIZE(x,WRAP2(y+1,hf->max_y),hf->max_x) ;
			z3 = *(hf->hf_buf+ii) ;
			if (z3>(value+threshold)) {
				totsup+=z3;
				nsup++;
				v[nsup-1] = ii;
				p[nsup-1]=z3;
			}
			ii=VECTORIZE(WRAP2(x-1,hf->max_x), WRAP2(y+1,hf->max_y),hf->max_x)  ;
			z4 = *(hf->hf_buf+ii);
			if (z4>(value+threshold)) {
				totsup+=z4;
				nsup++;
				v[nsup-1] = ii;
				p[nsup-1]=z4;
			}
			ii=VECTORIZE(WRAP2(x-1,hf->max_x),y,hf->max_x) ;
			z5 =  *(hf->hf_buf+ii) ;
			if (z5>(value+threshold)) {
				totsup+=z5;
				nsup++;
				v[nsup-1] = ii;
				p[nsup-1]=z5;
			}
			ii=VECTORIZE(WRAP2(x-1,hf->max_x),WRAP2(y-1,hf->max_y),hf->max_x) ;
			z6 = *(hf->hf_buf+ii);
			if (z6>(value+threshold)) {
				totsup+=z6;
				nsup++;
				v[nsup-1] = ii;
				p[nsup-1]=z6;
			}
			ii=VECTORIZE(x,WRAP2(y-1,hf->max_y),hf->max_x) ;
			z7 = *(hf->hf_buf+ii); 
			if (z7>(value+threshold)) {
				totsup+=z7;
				nsup++;
				v[nsup-1] = ii;
				p[nsup-1]=z7;
			}
			ii=VECTORIZE(WRAP2(x+1,hf->max_x),WRAP2(y-1,hf->max_y),hf->max_x) ;
			z8 = *(hf->hf_buf+ii);
			if (z8>(value+threshold)) {
				totsup+=z8;
				nsup++;
				v[nsup-1] = ii;
				p[nsup-1]=z8;
			}
			if (nsup) {

				*(deltamap+i) = *(deltamap+i) + 0.5* (totsup / nsup - value);
				for (vi=0; vi<nsup; vi++)
					*(deltamap+v[vi])= *(deltamap+v[vi])
						-0.5*( totsup/nsup - value ) * p[vi]/totsup;

//	For oriented erosion (we forget to sum the deltas...)

//				*(deltamap+i) = 0.5* (totsup / nsup - value);
//				for (vi=0; vi<nsup; vi++)
//					*(deltamap+v[vi])= 
//						-0.5*( totsup/nsup - value ) / nsup;
*
			}

		}
	}
	for (i=0; i<hf->max_x*hf->max_y; i++) {
		val = ((gint) *(hf->hf_buf+i)) + *(deltamap+i);
		*(hf->hf_buf+i) = MAX(MIN(val,0xFFFF),0);
	}
	}
// printf("MAX: %d; MIN: %d; DELTA: %d\n",hf->max, hf->min, hf->max-hf->min);
	free(deltamap);
}
 */
 
/*

gboolean find_min_neighbour( gint x, gint y, gint *newx, gint *newy,
	gint max, hf_type *hf, gint threshold) {
	//	Find the minimum neighbour of (x,y) in hf
	//	Returns the result in newx and newy
	//	If (x,y) is a minimum, returns FALSE
	static gint z1, z2, z3, z4, z5, z6, z7, z8, i, ii, nx, ny;
	hf_type lower;
	ii=VECTORIZE(WRAP2(x+1,max),y,max);
	z1 =  *(hf+ii);
	lower = z1;
	nx = WRAP2(x+1,max);
	ny = WRAP2(y+1,max);
	ii=VECTORIZE(nx, ny,max) ;
	z2 = *(hf+ii);
	if (z2<z1) {
		*newx = nx;
		*newy = ny;
		lower = z2;
	}
	ny=WRAP2(y+1,max) ;
	ii =VECTORIZE(x,ny,max) ;
	z3 = *(hf+ii) ;
	if (z3<lower) {
		*newx = x;
		*newy = ny;
		lower = z3;
	}
	nx = WRAP2(x-1,max);
	ny = WRAP2(y+1,max);
	ii=VECTORIZE(nx,ny,max)  ;
	z4 = *(hf+ii);
	if (z4<lower) {
		*newx = nx;
		*newy = ny;
		lower = z4;
	}
	nx = WRAP2(x-1,max) ;
	ii =VECTORIZE(nx,y,max) ;
	z5 =  *(hf+ii) ;
	if (z5<lower) {
		*newx = nx;
		*newy = y;
		lower = z5;
	}
	nx = WRAP2(x-1,max);
	ny = WRAP2(y-1,max);
	ii=VECTORIZE(nx, ny,max) ;
	z6 = *(hf+ii);
	if (z6<lower) {
		*newx = nx;
		*newy = ny;
		lower = z6;
	}
	ny = WRAP2(y-1,max) ;
	ii =VECTORIZE(x,ny,max) ;
	z7 = *(hf+ii);
	if (z7<lower) {
		*newx = x;
		*newy = ny;
		lower = z7;
	}
	nx = WRAP2(x+1,max);
	ny = WRAP2(y-1,max) ;
	ii=VECTORIZE(nx,ny,max)  ;
	z8 = *(hf+ii);
	if (z8<lower) {
		*newx = nx;
		*newy = ny;
		lower = z8;
	}
	if ( (lower+threshold) < *(hf + VECTORIZE(x,y,max)))
		return TRUE;
	else
		return FALSE;
}

*/
/*

void flow2(gint x, gint y, gint tox, gint toy, hf_type *hf, gint max, gint *water_weight) {
	//	Flow a quarter of the H difference between (x,y) and (tox,toy)
	//	from (x,y) to (tox,toy)
	//	The quantity flowed depends on the current water weight
//	Define a 3x3 matrix - proportion of the flowing chunk to displace
	static gint matrix[3][3]={1,2,1,2,4,2,1,2,1};
	gint i, toi, m, n, qty;
	hf_type nchunk;

	nchunk = *(hf+VECTORIZE(x,y,max)) - *(hf+VECTORIZE(tox,toy,max));
	nchunk = nchunk >> 4;
	for (m=0; m<3; m++)
		for (n=0; n<3; n++) {
			i = VECTORIZE(WRAP2(x-1+n,max),WRAP2(y-1+m,max),max);
			qty = nchunk * matrix[m][n];
			*(hf+i) = (hf_type) MIN(0xFFFF,MAX(0,(gint) *(hf+i) + qty));
		}
	for (m=0; m<3; m++)
		for (n=0; n<3; n++) {
			i = VECTORIZE(WRAP2(tox-1+n,max),WRAP2(toy-1+m,max),max);
			qty = nchunk * matrix[m][n];
			*(hf+i) = (hf_type) MIN(0xFFFF,MAX(0,(gint) *(hf+i) + qty));
		}

}
*/
/*

gboolean find_all_neighbours( gint x, gint y, gint *newx, gint *newy,
	gint max, hf_type *hf, gint threshold) {
	//	Find all the inferior neighbours of (x,y) in hf
	//	Randomly select one of the neighbours,
	//	given its height difference with the minimum is < to threshold
	//	Returns the result in newx and newy
	//	If (x,y) is a minimum, returns FALSE
	static gint v[8],z1, z2, z3, z4, z5, z6, z7, z8, i, ii, nx, ny, ninf, ok;
	long int value;
	long int lower= (long int) 0xFFFF;
	ninf = 0;
	ok = 0;
	value = *(hf+VECTORIZE(x,y,max)) ;
	ii=VECTORIZE(WRAP2(x+1,max),y,max);
	z1 =  *(hf+ii);
	if (z1<value) {
		v[ninf] = ii;
		ninf++;
	}
	lower = z1;
	nx = WRAP2(x+1,max);
	ny = WRAP2(y+1,max);
	ii=VECTORIZE(nx, ny,max) ;
	z2 = *(hf+ii);
	if (z2<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z2<z1)
		lower = z2;
	ny=WRAP2(y+1,max) ;
	ii =VECTORIZE(x,ny,max) ;
	z3 = *(hf+ii) ;
	if (z3<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z3<lower)
		lower = z3;
	nx = WRAP2(x-1,max);
	ny = WRAP2(y+1,max);
	ii=VECTORIZE(nx,ny,max)  ;
	z4 = *(hf+ii);
	if (z4<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z4<lower)
		lower = z4;
	nx = WRAP2(x-1,max) ;
	ii =VECTORIZE(nx,y,max) ;
	z5 =  *(hf+ii) ;
	if (z5<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z5<lower)
		lower = z5;
	nx = WRAP2(x-1,max);
	ny = WRAP2(y-1,max);
	ii=VECTORIZE(nx, ny,max) ;
	z6 = *(hf+ii);
	if (z6<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z6<lower)
		lower = z6;
	ny = WRAP2(y-1,max) ;
	ii =VECTORIZE(x,ny,max) ;
	z7 = *(hf+ii);
	if (z7<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z7<lower)
		lower = z7;
	nx = WRAP2(x+1,max);
	ny = WRAP2(y-1,max) ;
	ii=VECTORIZE(nx,ny,max)  ;
	z8 = *(hf+ii);
	if (z8<value) {
		v[ninf] = ii;
		ninf++;
	}
	if (z8<lower)
		lower = z8;

// printf("(X,Y): (%d,%d); ninf: %d; lower: %u; threshold: %d; value: %d; ",x,y,ninf, lower, threshold, value);
	if ( (lower+threshold) < value) {
		// Find all the pixels not farther away than "threshold" than the lower one
		for (i=0; i<ninf; i++) { 
			if ((*(hf+v[i])-threshold)<=lower) {
				v[ok] = v[i];
				ok ++;
			}
		}
// printf("ok: %d; ",ok);
		ii = v[rand()%ok];
// printf("ii: %d; ",ii);
		*newx = ii%max;
		*newy = (gint) (ii / max);
// printf("(NEWX,NEWY) = (%d,%d) = (%d,%d)\n", ii%max,(gint) ii/max,*newx,*newy);
		return TRUE;
	}
	else {
// printf("Returning FALSE\n");
		return FALSE;
	}
}
*/

/*
void hf_crests_erosion (hf_struct_type *hf, gint steps, gint thresh) {
//	Simple iterative erosion, giving crested mountains
//	We find the minimum neighbour
//	If its value is superior to the current pixel one,
//	we add half the difference to the current pixel
//	If the minimum neighbour is inferior, we subtract half the current pixel
//	This is like if half the difference was "flowing" from or to the current pixel
//	Neighbours are numbered from 1 to 8, clockwise, 1 being (1,0)
//	2004-03-11 Corrected to deal with wrapping
	static gint z1, z2, z3, z4, z5, z6, z7, z8;
	static gint x,y, i, s, val, ii, minii, lower, threshold;
	hf_type value;
	gfloat ratio; // Factor for dividing the weight of the minimum pixel, when it is on the diagonals
	threshold = (gint) pow(2.0,(double) thresh);
	if (!hf->tmp_buf)
		hf_backup(hf);
	// We need a second temporary buffer for storing
	// the result of each step
	if (!hf->result_buf) {
		hf->result_buf = (hf_type *)malloc(hf->max_x*hf->max_y*sizeof(hf_type));
	}
// printf("CRESTS EROSION, STEPS: %d; THRESHOLD: %d\n", steps, threshold);
	memcpy(hf->result_buf, hf->tmp_buf, hf->max_x*hf->max_y*sizeof(hf_type));
	for (s=0; s<steps; s++) {
	for (y=0; y<hf->max_y; y++) {
		for (x=0; x<hf->max_x; x++) {
			i = VECTORIZE(x,y,hf->max_x);
			value = *(hf->hf_buf + i);
			if (hf->if_tiles)
				ii=VECTORIZE(WRAP2(x+1,hf->max_x),y,hf->max_x);
			else
				ii=VECTORIZE(MIN(x+1,hf->max_x-1),y,hf->max_x);
			z1 =  *(hf->hf_buf+ii);
			minii = ii;
			lower = z1;
			ratio = 1.0;
			if (hf->if_tiles)
				ii=VECTORIZE(WRAP2(x+1,hf->max_x), WRAP2(y+1,hf->max_y),hf->max_x) ;
			else
				ii=VECTORIZE(MIN(x+1,hf->max_x-1), MIN(y+1,hf->max_y-1),hf->max_x) ;
			z2 = *(hf->hf_buf+ii);
			if (z2<z1) {
				minii = ii;
				lower = z2;
				ratio = 1.4142;
			}
			if (hf->if_tiles)
				ii =VECTORIZE(x,WRAP2(y+1,hf->max_y),hf->max_x) ;
			else
				ii =VECTORIZE(x,MIN(y+1,hf->max_y-1),hf->max_x) ;
			z3 = *(hf->hf_buf+ii) ;
			if (z3<lower) {
				minii = ii;
				lower = z3;
				ratio = 1.0;
			}
			if (hf->if_tiles)
				ii=VECTORIZE(WRAP2(x-1,hf->max_x), WRAP2(y+1,hf->max_y),hf->max_x)  ;
			else
				ii=VECTORIZE(MAX(x-1,0), MIN(y+1,hf->max_y-1),hf->max_x)  ;
			z4 = *(hf->hf_buf+ii);
			if (z4<lower) {
				minii = ii;
				lower = z4;
				ratio = 1.4142;
			}
			if (hf->if_tiles)
				ii=VECTORIZE(WRAP2(x-1,hf->max_x),y,hf->max_x) ;
			else
				ii=VECTORIZE(MAX(x-1,0),y,hf->max_x) ;
			z5 =  *(hf->hf_buf+ii) ;
			if (z5<lower) {
				minii = ii;
				lower = z5;
				ratio = 1.0;
			}
			if (hf->if_tiles)
				ii=VECTORIZE(WRAP2(x-1,hf->max_x),WRAP2(y-1,hf->max_y),hf->max_x) ;
			else
				ii=VECTORIZE(MAX(x-1,0),MAX(y-1,0),hf->max_x) ;
			z6 = *(hf->hf_buf+ii);
			if (z6<lower) {
				minii = ii;
				lower = z6;
				ratio = 1.4142;
			}
			if (hf->if_tiles)
				ii=VECTORIZE(x,WRAP2(y-1,hf->max_y),hf->max_x) ;
			else
				ii=VECTORIZE(x,MAX(y-1,0),hf->max_x) ;
			z7 = *(hf->hf_buf+ii);
			if (z7<lower) {
				minii = ii;
				lower = z7;
				ratio = 1.0;
			}
			if (hf->if_tiles)
				ii=VECTORIZE(WRAP2(x+1,hf->max_x),WRAP2(y-1,hf->max_y),hf->max_x) ;
			else
				ii=VECTORIZE(MIN(x+1,hf->max_x-1),MAX(y-1,0),hf->max_x) ;
			z8 = *(hf->hf_buf+ii);
			if (z8<lower) {
				minii = ii;
				lower = z8;
				ratio = 1.4142;
			}
			val = ( ( ((gint) *(hf->hf_buf+minii)) - (gint) value ) / 2);
			if (threshold < ABS(val))
				*(hf->result_buf+i) = value + (gint) ( ((gfloat) val) / ratio);
			else
				*(hf->result_buf+i) = value;
		}
	}
	memcpy(hf->hf_buf, hf->result_buf, hf->max_x*hf->max_y*sizeof(hf_type));
	}
// printf("MAX: %d; MIN: %d; DELTA: %d\n",hf->max, hf->min, hf->max-hf->min);
}

/*
void hf_rain_erosion2 (hf_struct_type *hf, gint dr_log, gint thresh) {
//	Simple iterative erosion
//	We generate random drops and follow their path
//	Neighbours are numbered from 1 to 8, counterclockwise, 1 being (1,0)
//	The default for this version of the program is to wrap
	static gint x,y, i, ii, s, val, threshold, nextx, nexty, lngpath;
	gint *water_weight;
	unsigned long int drops, steps;
//	threshold is the future chunk... it must be at least 16, because the center
//	of the 3x3 averaging matrix == 4 and we divide the chunk by 16 before applying it
	threshold = 15 + (gint) pow(2.0,(double) thresh);
	drops = (long int) pow(2.0,(gdouble) dr_log);
	water_weight = (gint *) calloc(hf->max_x*hf->max_y,sizeof(gint));
	if (!hf->tmp_buf)
		hf_backup(hf);
printf("RAIN EROSION, DROPS: %d; THRESHOLD: %d\n", drops, threshold);
	for (s=0; s<drops; s++) {
		x = rand()%hf->max_x;
		y = rand()%hf->max_y;
		// Make the (x,y) drop flow until it encounters a minimum
		steps = 0;
// printf("********** DROP: %u; (x,y): (%d,%d)\n",s,x,y);
		while (find_all_neighbours(x,y,&nextx,&nexty,
				hf->max_x,hf->hf_buf,threshold,
				hf->if_tiles))	{
			flow(x,y,nextx,nexty,hf->hf_buf,hf->max_x, water_weight);
			ii = VECTORIZE(nextx,nexty,hf->max_x);
			y = nexty;
			x = nextx;
			steps++;
			if (steps>hf->max_x)
				break;
		}
	}
	free(water_weight);
}

*/
/*
void hf_rain_erosion (hf_struct_type *hf, gint drops, gint thresh) {
//	Simple iterative erosion
//	We generate random drops and follow their path
//	We don't flow a drop when the slope is more than 2 power thresh
//	Neighbours are numbered from 1 to 8, clockwise, 1 being (1,0)
//	Not extensively tested when the HF doesn't wrap
	static gint x,y, i, ii, s, val, threshold, nextx, nexty, lngpath;
	unsigned long int steps;
//	threshold is the future chunk... it must be at least 16, because the center
//	of the 3x3 averaging matrix == 4 and we divide the chunk by 16 before applying it
	threshold = 15 + (gint) pow(2.0,(double) thresh);
//	In the 1st version, drops were given as an exponent
//	We could fallback to this in the future...
//	drops = (long int) pow(2.0,(gdouble) dr_log);
	if (!hf->tmp_buf)
		hf_backup(hf);
// printf("RAIN EROSION, DROPS: %d; THRESHOLD: %d\n", drops, threshold);
	for (s=0; s<drops; s++) {
		x = rand()%hf->max_x;
		y = rand()%hf->max_y;
		// Make the (x,y) drop flow until it encounters a minimum
		steps = 0;
// printf("********** DROP: %u; (x,y): (%d,%d)\n",s,x,y);
		while (find_all_neighbours(x,y,&nextx,&nexty,
				hf->max_x,hf->hf_buf,threshold))	{
			if (!hf->if_tiles)
				if ( (x==0) || (y==0) ||
					(x==(hf->max_x-1)) ||
					(y==(hf->max_y-1)) )
						break;
			flow(x,y,nextx,nexty,hf->hf_buf,hf->max_x, hf->if_tiles);
			ii = VECTORIZE(nextx,nexty,hf->max_x);
			y = nexty;
			x = nextx;
			steps++;
			if (steps>hf->max_x)
				break;
		}
	}
}
*/

/*
// No improvement with hexagonal sampling, gives annoying artifacst (lines aliasing)
void hf_eroded_ribs_hex (hf_struct_type *hf) {
//	Build a "ribbed" (eroded) template by taking the minimum
//	absolute value of the difference between each pixel and its neighborus
//	Neighbours are numbered from 1 to 8, clockwise, 1 being (1,0)
//	Version with hexagonal sampling
//	The default for this version of the program is to wrap
	static gint z1, z2, z3, z4, z5, z6, shift;
	hf_type value, *hf_in, *hf_out;
	gboolean test, even=TRUE;
	static gint x,y, i, val;
	gint max;
	
	max = hf->max_x;		

	hf_in = hexagonal_row_sampling (hf->hf_buf, max, 
		TRUE, even, TRUE);		
	hf_out = hf->result_buf;
			
	test = y & (gint) 1; // test = TRUE if value is odd
	if ( (test && even) || ((!test) && (!even)) )  //  XOR
		shift = 1;
	else
		shift = 0;
		
	for (y=0; y<max; y++)
		for (x=0; x<max; x++) {
			i = VECTORIZE(x,y,max);
			value = *(hf_in + i);
			z1 =  *(hf_in+VECTORIZE(WRAP2(x+1,max),y,max))  ;
			z2 = *(hf_in+VECTORIZE(WRAP2(x+shift,max),
				WRAP2(y+1,max),max)) ;
			z3 = *(hf_in+VECTORIZE(WRAP2(x-1+shift,max),WRAP2(y+1,max),max))  ;
			z4 = *(hf_in+VECTORIZE(WRAP2(x-1,max),
				WRAP2(y+1,max),max))  ;
			z5 =  *(hf_in+VECTORIZE(WRAP2(x-1+shift,max),WRAP2(y-1,max),max)) ;
			z6 = *(hf_in+VECTORIZE(WRAP2(x+shift,max),
				WRAP2(y-1,max),max)) ;

//			Absolute minimum (negative of)
			z1 = ABS(z1-value);
			z2 = ABS(z2-value);
			z3 = ABS(z3-value);
			z4 = ABS(z4-value);
			z5 = ABS(z5-value);
			z6 = ABS(z6-value);

			*(hf_out + i)  = ~ (hf_type) MIN(z1,MIN(z2,MIN(z3,MIN(z4,MIN(z5,z6))))) ;
// if (!(i%512)) {
// printf("*(hf_out+i): %d; z1: %d, z2: %d, z3: %d, z4: %d, z5: %d, z6: %d, z7: %d, z8: %d\n",*(hf_out+i), z1, z2, z3, z4, z5, z6, z7, z8);
// }
		}		
	// Restore the rectangular sampled hf
	hf->result_buf = hexagonal_row_sampling (hf_out, max, TRUE, even, FALSE); 
	free(hf_out);
	free(hf_in);
}
*/
