/* img_process.c - functions for filtering and transforming images
 *
 * Copyright (C) 2001 Patrice St-Gelais
 *         patrstg@users.sourceforge.net
 *         www.oricom.ca/patrice.st-gelais
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "img_process.h"
#include "hf.h"
#include "hf_filters.h"

void hf_auto_contrast (hf_struct_type *hf) {
//	Stretch contrast so that the full value range is used (0 to 0xFFFF)
	hf_clamp (hf, 0, 0xFFFF);
}

void hf_brightness_contrast(hf_struct_type *hf, gint contrast_level, gint br_level,
		gboolean keep_luminosity, gboolean bound_overflow) {
//	
//	Increases / decreases the contrast & brightness of a square size X size 16 greyscale image
//	Levels are a % from -100 to 100
//	The brightness process is additive
//	At -100%, we subtract 0xFFFF>>1;  at 100%, we add 0xFFFF>>1.
//	The contrast is relative to the medium value between the min and max
//	The input is hf->tmp_buf (backup of the displayed buffer before getting into
//	the contrast dialog) and the output is the displayed buffer hf->hf_buf
//	Min and Max should be updated BEFORE, when entering into the dialog

//	The contrast process adjusts the delta between min and max value, so that
//	the new delta is 0 at -100% level and 2x at 100% level

	float flevel;
	gint i, max = (gint) 0xFFFF, max2 , value, offset, delta_brightness ;
	max2 = 2 * max - 1;
	if (!hf->tmp_buf)
		hf_backup(hf);
	delta_brightness =  (gint) ( (0xFFFF>>1) * ((gfloat) br_level / 100));
//	flevel ranges from 0.0 to 2.0
	flevel =  ((gfloat) (100 + contrast_level)) / 100 ;
	if (keep_luminosity)
		offset = hf->avrg * (1-flevel);
	else
		offset = 0;
	for (i=0; i<hf->max_x*hf->max_y; i++) {
		//	If contrast_level < 0, we apply brightness 2nd
		if (contrast_level < 0) {
			if (bound_overflow) {
			//	Contrast
				*(hf->hf_buf + i) = MAX(0,MIN(0xFFFF, offset 
					+ ((gfloat) *(hf->tmp_buf + i)) * flevel));
			//	Brightness
				if (br_level)
					*(hf->hf_buf + i) = 
						(hf_type) MIN(0xFFFF,MAX(0,
						(gint) *(hf->hf_buf + i) + delta_brightness));
			}
			else { // Reverted overflow
			// 	Contrast
				value = offset + ((gfloat) *(hf->tmp_buf+i)) * flevel;
			// 	Brightness
				if (br_level)
					value = value + delta_brightness;
				value = ABS(value);
				value = (value>=max)?(max2-value):value;
				*(hf->hf_buf+i) = (hf_type) value;
			}
		}
		else {
			if (bound_overflow) {
			//	Brightness
				*(hf->hf_buf + i) = 
					(hf_type) MIN(0xFFFF,MAX(0,
						(gint) *(hf->tmp_buf + i) + delta_brightness));
			//	Contrast
				if (contrast_level)
					*(hf->hf_buf + i) = MAX(0,MIN(0xFFFF, offset 
						+ ((gfloat) *(hf->hf_buf + i)) * flevel));
			}
			else { // Reverted overflow
			// 	Brightness
				value =  *(hf->tmp_buf+i) + delta_brightness;
			// 	Contrast
				if (contrast_level)
					value = offset + ((gfloat) value) * flevel;
				value = ABS(value);
				value = (value>=max)?(max2-value):value;
				*(hf->hf_buf+i) = (hf_type) value;
			}
		}
	}
}

void hf_remove_artifacts (hf_struct_type *hf, gint radius, dist_matrix_struct *dm) {
	// Removing artifacts in a quantized image (terrace)
	// Move a window of a given radius over the image
	// and fill it with the border color if all the border pixels are the same
	// In the simplest form, the window is square
	// The function "wraps"
	gint i, j, k, l;
	hf_type value;
	gboolean test;
	for (i=0; i<hf->max_x; i++) {
		for (j=0; j<hf->max_y; j++) {
			value = *(hf->hf_buf + i + j*hf->max_x);
			test = TRUE;
			for (k=-radius; k<radius+1; k++) {
				if (value != *(hf->hf_buf + WRAP2(i - radius,hf->max_x)
					 + WRAP2((j+k),hf->max_y) *hf->max_x)) {
					test = FALSE;
					break;
				}
				if (value != *(hf->hf_buf + WRAP2(i + radius,hf->max_x)
					 + WRAP2((j+k),hf->max_y) *hf->max_x)) {
					test = FALSE;
					break;
				}
			} // End for k (1st)
			if (!test)
				continue;
			for (k = -radius+1; k<radius; k++) {
				if (value != *(hf->hf_buf + WRAP2(i + k,hf->max_x)
					 + WRAP2((j-radius),hf->max_y)*hf->max_x)) {
					test = FALSE;
					break;
				}
				if (value != *(hf->hf_buf + WRAP2(i + k,hf->max_x)
					 + WRAP2((j+radius),hf->max_y)*hf->max_x)) {
					test = FALSE;
					break;
				}
			} // End for k (2nd)
 			if (!test)
				continue;
			// We fill up the window with "value"
			for (k=-radius+1; k<radius; k++)
				for (l=-radius+1; l<radius; l++)
					*(hf->hf_buf+VECTORIZE(
						WRAP2(i+l,hf->max_x),
						WRAP2(j+k,hf->max_y),
							hf->max_x) )= value;
					
		} // Enf for j
	} // End for i
}

void hf_terrace (hf_struct_type *hf, gint levels, gint mix, gint seed, gint percent_random,
	gint merge_op, gint source_offset, gint result_offset, 
	gint smooth_radius, gboolean if_wrap, 
	dist_matrix_struct *dist_matrix,
	gint artifact_radius, gboolean apply_postprocess,
	gboolean merge_change,
	gint merge_x_translate,
	gint merge_y_translate,
	gdouble **gauss_list) {

	gint i,*disturbance_vector, index, slice, index_offset;
	hf_type hf_source, hf_result;
	gfloat source, result, s_offset, r_offset;
	static gfloat max_f = (gfloat) MAX_HF_VALUE;

	if (!levels)
		return;
	slice = (hf->max-hf->min) / levels;
	//	No process if slice is 0 - HF is uniform
	if (!slice)
		return;
	// Modified 2005-06-23: s_offset and r_offset are now %,not absolute values
	/*
	s_offset = (slice* source_offset) / max_f;	// source_offset and result_offset are % of slice
	r_offset = (slice* result_offset) / max_f ;
printf("slice: %d; s_offset: %f;  r_offset: %f\n",slice, s_offset, r_offset);
	*/
	if (!hf->tmp_buf)
		hf_backup(hf);
	// We need a second temporary buffer for storing
	// the result of the quantization process before the postprocessing
	if (!hf->result_buf) {
		hf->result_buf = (hf_type *)malloc(hf->max_x*hf->max_y*sizeof(hf_type));
	}
//	Variation implementing random disturbance of each level
//	Disturbance is bounded by 0 and slice_size
//	(NOTE: the terrace effect is calculating by "flooring" a value
//	to the nearest lower multiple of slice_size, so the total H
//	should never be more than 0xFFFF)
/*	disturbance_vector = (gint *) malloc(sizeof(gint)*(levels+1));
	srand(seed);
	*(disturbance_vector) = 0;	// We want the "floor" to stay black
	for (i=1; i<levels+1; i++) {
		*(disturbance_vector+i) = (percent_random*(rand()%slice)) / 100;
	}
	for (i=0; i<hf->max_x*hf->max_y; i++) {
		index = *(hf->tmp_buf + i) / slice;
		*(hf->hf_buf + i) = ((gfloat) *(hf->tmp_buf +i)* (100-percent) / 100) +
			((gfloat) ( (*(disturbance_vector+index) +
			(slice * index) )* percent) / 100);
	}
*/
//	2nd variation
//	The disturbance vector is applied when quantizing, not after
//	An offset to the merge function is added
//	Merge can now be done with +, -, *, min or max
	disturbance_vector = (gint *) malloc(sizeof(gint)*(levels+1));
	srand(seed);
	*(disturbance_vector) = hf->min;	// We want the "floor" to stay at minimum

	for (i=1; i<levels+1; i++) {
		*(disturbance_vector+i) = (percent_random*(rand()%slice)) / 100;
		// 	For the quantization algorithm to work, 
		// 	each slice displacement must be strictly < to slice_size (never =)
		*(disturbance_vector+i) = MIN(-1+((i+1)*slice), *(disturbance_vector+i));
		// 	Contrary to the simpler preceding algorithm,
		// 	the quantization requires cumulative values, so that we can
		// 	rapidly check upper and lower bounds
		*(disturbance_vector+i) = MIN(0xFFFF,hf->min + (i*slice)+*(disturbance_vector+i));
// printf("DIST-VECT[%d] = %d\n",i,*(disturbance_vector+i));
	}

//	When the minimum HF value is > slice size, we need to adjust the index
	index_offset = hf->min / slice;
//	If a smooth radius or an artifact removal radius is given, 
//	we must decompose the process in 2 passes:
   if ( apply_postprocess && (smooth_radius || artifact_radius) ) {
	for (i=0; i<hf->max_x*hf->max_y; i++) {
		// 	First, we approximate quantization
		index = (*(hf->tmp_buf + i) / slice) - index_offset;
		// 	"index*slice_size" gives the 1st value to test
		// 	Since displacement is strictly less than the slice_size,
		// 	the looked for value can only be index or index-1
		if (*(disturbance_vector+index)>*(hf->tmp_buf+i))
			index = MAX(0,index-1);  // This MAX is not supposed to be required!!
		*(hf->hf_buf + i) = *(disturbance_vector+index);
	}
	// 	The 2 next processes operate on hf->hf_buf
	if (artifact_radius)
		hf_remove_artifacts(hf, artifact_radius, dist_matrix);
	if (smooth_radius) 
		hf_smooth(hf, smooth_radius, dist_matrix, if_wrap, gauss_list);
	// 	Swap result_buf and hf_buf, so that result_buf becomes the input
	// 	and hf_buf is prepared for a new output
	swap_buffers ((gpointer*) &hf->result_buf, (gpointer*) &hf->hf_buf);
  }	// 	end of "if (apply_postprocess)"
  else  {	// 	if ( ! apply_postprocess)
	// 	If (merge_change), we are only re-merging the result and the source
	// 	Otherwise, the result is recalculated in result_buf
	if (!merge_change) {
		// Note:  we could merge hf_simple_merge with this code for avoiding 1 loop
		// However, the repeated code from hf_simple_merge would be more difficult to maintain
		// One alternate solution would be to "embed" a function in hf_simple_merge
		for (i=0; i<hf->max_x*hf->max_y; i++) {
			index =  (*(hf->tmp_buf + i) / slice) - index_offset;
			if (*(disturbance_vector+index)>*(hf->tmp_buf+i))
				index = MAX(0,index-1);  
			*(hf->result_buf + i) = *(disturbance_vector+index);
		}
	}
  }	
  hf_simple_merge_translate (hf->tmp_buf, hf->result_buf, hf->hf_buf, mix, source_offset, result_offset, 
		hf->max_x, hf->max_y, merge_op, merge_x_translate, merge_y_translate);		
  if (disturbance_vector)
	free(disturbance_vector);
}

void hf_threshold (hf_struct_type *hf, hf_type min, hf_type max, gint percent) {
	gint i;
	if (!hf->tmp_buf)
		hf_backup(hf);
	for (i=0; i<hf->max_x*hf->max_y; i++) {
		*(hf->hf_buf + i) = (hf_type) ((long int) *(hf->tmp_buf +i)* (100-percent) / 100) +
			( ( (long int) ( MAX(min,MIN(max,*(hf->tmp_buf + i))) ) * percent) / 100);
	}
}

void hf_convolve (hf_struct_type *hf, gint radius, gdouble *matrix, gboolean if_tiles) {
//	Apply a convolution filter
//	matrix size = (2*radius+1)*(2*radius+1)
	gint i,j,k,l, ii, x0,y0,width;
	gdouble value,weight=0.0;
	hf_type *hf_buf;
	gboolean if_weight = TRUE;
	width = 2*radius + 1;
	hf_buf = (hf_type *) malloc(sizeof(hf_type)*hf->max_x*hf->max_y);
	for (i=0;i<width*width;i++) {
		weight += *(matrix+i);
	}
	if ((weight==0.0) || (weight==1.0))
		if_weight = FALSE;
// printf("W: %s... %f\n",if_weight?"TRUE":"FALSE",weight);
	for (i=0; i<hf->max_y; i++) {
		ii = i*hf->max_x;
		for (j=0; j<hf->max_x; j++) {
			value = 0.0;
			y0 = i - radius;
			x0 = j - radius;
		//	Add the values contained in the (2*radius)+1 *... square centered at current dot
		  	 for (k = 0; k < width; k++)
				for (l = 0; l < width; l++) {
					if ((*matrix+k*width+l)>0.0)
						if (if_tiles)
			 			   value = value + *(matrix+k*width+l) * 
					 	     (gdouble) *(hf->hf_buf + 
						     WRAP2(y0+k,hf->max_y)*(hf->max_y) +
						     WRAP2(x0+l,hf->max_x)) ;
						else
			 			   value = value + *(matrix+k*width+l) *
					 	      (gdouble) *(hf->hf_buf  +
						     MAX(0,MIN(y0+k,hf->max_y-1))*(hf->max_y) +
						     MAX(0,MIN(x0+l,hf->max_x-1)) )  ;
				}
			if (if_weight)
				value = value / weight;
			*(hf_buf + ii+j) = (hf_type) MAX(0.0,value) ;
		}
	}
	memcpy(hf->hf_buf,hf_buf,sizeof(hf_type)*hf->max_x*hf->max_y);
	free(hf_buf);
}

void hf_convolveb (hf_struct_type *hf, gint radius, gdouble *matrix, gboolean if_tiles) {
//	"Boolean" convolution filter
//	matrix size = (2*radius+1)*(2*radius+1)
//	If the current pixel is different from any neighbour, put it black
	gint i,j,k,l, ii, x0,y0,width;
	gboolean test;
	gdouble weight=0.0;
	hf_type *hf_buf, value;
	gboolean if_weight = TRUE;
	width = 2*radius + 1;
	hf_buf = (hf_type *) malloc(sizeof(hf_type)*hf->max_x*hf->max_y);
	for (i=0;i<width*width;i++) {
		weight += *(matrix+i);
	}
	if ((weight==0.0) || (weight==1.0))
		if_weight = FALSE;
// printf("W: %s... %f\n",if_weight?"TRUE":"FALSE",weight);
	for (i=0; i<hf->max_y; i++) {
		ii = i*hf->max_x;
		value = *(hf->hf_buf + ii+j) ;
		for (j=0; j<hf->max_x; j++) {
			test = TRUE;
			y0 = i - radius;
			x0 = j - radius;
		//	Add the values contained in the (2*radius)+1 *... square centered at current dot
		  	 for (k = 0; k < width; k++)
				for (l = 0; l < width; l++) {
					if ((*matrix+k*width+l)>0.0)
						if (if_tiles)
			 			   test = test && (value == 
					 	     *(hf->hf_buf + 
						     WRAP2(y0+k,hf->max_y)*(hf->max_y) +
						     WRAP2(x0+l,hf->max_x)) ) ;
						else
			 			   test = test && (value == 
					 	      *(hf->hf_buf  +
						     MAX(0,MIN(y0+k,hf->max_y-1))*(hf->max_y) +
						     MAX(0,MIN(x0+l,hf->max_x-1)) )  );
				}
			if (!test)
				*(hf_buf + ii+j) = 0 ;
			else
				*(hf_buf + ii+j) = *(hf->hf_buf+ii+j) ;
		}
	}
	memcpy(hf->hf_buf,hf_buf,sizeof(hf_type)*hf->max_x*hf->max_y);
	free(hf_buf);
}

//	static gdouble matrix[] = {0.0625,0.125,0.0625,0.125,0.25,0.125,0.0625,0.125,0.0625}; // Blur
//	"Sharpen" / add noise;  without clamping and with sum = 0:  B&W edge detection
//	static gdouble matrix[] = {-0.5,-0.5,-0.5,-0.5,4.0,-0.5,-0.5,-0.5,-0.5}; 
//	static gdouble matrix[] = {1,1,1,1,1,1,1,1,1}; // For min/max (erode/dilate)
//	static gdouble matrix[] = {-1,-1,-1,-1,8.0,-1,-1,-1,-1}; 

void hf_erode (hf_struct_type *hf, gint steps) {
//	Simple erosion with a modified min/max convolution filter
//	matrix size = (2*radius+1)*(2*radius+1)
//	static gdouble matrix[] = {0.7,1,0.7,1,1,1,0.7,1,0.7};
	static gdouble matrix[] = {0,1,0,1,1,1,0,1,0};
//	static gdouble matrix[] = {1,1,1,1,1,1,1,1,1};
	gint i,j,k,l, ii, s, x0,y0,width, radius = 1, totsize;
	gdouble value,last,current, weight;
	hf_type *hf_in, *hf_out;
	width = 2*radius + 1;
	for (s=1; s<steps; s++) {
	   totsize = sizeof(hf_type) * hf->max_x *hf->max_y;
	   hf_in = (hf_type *) malloc(totsize);
	   memcpy (hf_in,hf->hf_buf, totsize);
  	   for (i=0; i<hf->max_y; i++) {
		ii = i*hf->max_x;
		for (j=0; j<hf->max_x; j++) {
			current = (gdouble) *(hf_in + ii+j) ;
			value = current;
			y0 = i - radius;
			x0 = j - radius;
		//	Add the values contained in the (2*radius)+1 *... square centered at current dot
		  	 for (k = 0; k < width; k++)
				for (l = 0; l < width; l++) {
					weight = *(matrix+k*width+l);
					if (weight>0.0) 
								 {
//	Minimum normalized with matrix weight (typically distance)
						last = 	(gdouble) *(hf_in + 
							MAX(0,MIN(y0+k,hf->max_y-1))*(hf->max_y) +
							MAX(0,MIN(x0+l,hf->max_x-1)) );
						if (last<current)
							value = (1.0-weight)*current+weight*last;
					}
/*
					 value = MIN(value,
						(gdouble) *(hf_in + 
						MAX(0,MIN(y0+k,hf->max_y-1))*(hf->max_y) +
						MAX(0,MIN(x0+l,hf->max_x-1)) ) );
*/
				}
			*(hf->hf_buf + ii+j) = (hf_type) value ;
		} // j (x) loop
	   } // i (y) loop
	   free(hf_in);
	} // s (steps) loop
}

void hf_map_convolve (hf_struct_type *map, hf_struct_type *background, gint cx, gint cy,
	gboolean wrap, gint level, gdouble **gauss_list) {

	map_convolve (map->hf_buf, map->max_x, map->max_y,
		background->hf_buf, background->max_x, background->max_y, 
		cx, cy, wrap, level, gauss_list, FALSE);
}

hf_type* init_blocks (hf_type *hf, gint max_x, gint max_y, gint radius, gint nblocks,
		gfloat phase, gboolean if_tiles) {
//	Average the pixels of hf into a smaller nblocks * nblocks hf_buf
//	Kind of "undersampling";  basis for processes as "smooth" and "cityscape"
	gint i, j, k, l, x, y;
	hf_type *blocks;
	long int value;
	blocks = (hf_type *) malloc(nblocks * nblocks * sizeof(hf_type));
	for (i=0; i<nblocks; i++) {
  	  for (j=0; j<nblocks; j++) {
		value = 0;
		for (k=((i+phase)*radius); k<((i+1+phase)*radius); k++)
		for (l=((j+phase)*radius); l<((j+1+phase)*radius); l++) {
			if (if_tiles) {
				x = (max_x+l)%max_x;
				y = (max_y+k)%max_y;
			}
			else {
				x = MIN(l, max_x-1);
				y = MIN(k, max_y-1);
			}
			value = value + *(hf + y*max_x + x);
		}	// End k and l loops
		*(blocks+(i*nblocks)+j) = (hf_type) (value / (radius * radius));
// printf("BLOCS: %d;  BLOC(%d,%d) = %d;  value = %d\n",nblocks,i,j,*(blocks+(i*radius)+j),value );
	  }	// End j loop
	}	// End i loop
	return blocks;
}

void hf_smooth (hf_struct_type *hf, gint radius,
	dist_matrix_struct *dm, gboolean if_tiles,gdouble **gauss_list) {
//	Smoothing process, based on a sampling, followed by a rebuild 
//	based on a normal (gaussian) 3d function (used as "shape filters" elsewhere in Geomorph)
//	For radius <= 36, we use a "separated" convolution in the spacial domain
//
//	dist_matrix = the usual precompiled distances matrix, for managing gaussian filters
//	if_tiles = TRUE => "wraps":  we reuse pixel 0 after pixel max_x or max_y

	gint i, j, k, l, nblocks, box, index, indexf, offset, half_radius;
	hf_type *blocks;
	filter_struct *filter;
	long int value;
	gfloat fvalue;
	gfloat *hf_weight,*hf_values;
	gdouble *v;

	if (!radius)	// We are not supposed to be here!
		return;
	if (!hf->tmp_buf)
		hf_backup(hf);

	if (radius<=36) {
		if (!gauss_list[radius])
			gauss_list[radius] = normalized_bell_new(radius);
		v = gauss_list[radius];
		convolve_normalized_vector (hf->hf_buf,
			hf->hf_buf, 
			hf->max_x, 
			hf->max_y, 
			if_tiles,
			radius, 
			v,
			HF_TYPE_ID);
		}
	else {

//	We add repeatedly a gaussian bell, at each sampled pixel
//	2003-02-14:  finally there is only 1 level left, since a separable convolution is used
//	Finally the best rebuilding box radius is equal to the specified radius
	box = radius;
	filter = sharp_filter_new(box<<1,dm);	// The size must be even
	half_radius = radius>>1;

//	1- Sample
	nblocks = (gint) ((hf->max_x-1+half_radius) / half_radius);
// printf("NBLOCKS: %d\n",nblocks);
	blocks = init_blocks(hf->hf_buf, hf->max_x, hf->max_y, half_radius, nblocks, 0, if_tiles);

	hf_values = (gfloat *) calloc(hf->max_x*hf->max_y,sizeof(gfloat));
	hf_weight = (gfloat *) calloc(hf->max_x*hf->max_y, sizeof(gfloat));

//	2- Rebuild
	// The sampling / rebuild process introduces a translation, which we must correct
	offset = half_radius/2;
	for (i=0; i<nblocks; i++) {	
  	  for (j=0; j<nblocks; j++) {	
		
		for (k=0; k<filter->hf_size; k++)
			for (l=0; l<filter->hf_size; l++) {

				if (if_tiles)
				   index = WRAP2((i*half_radius+k-box+offset),hf->max_y)*hf->max_x
						 + WRAP2(j*half_radius +l-box+offset,hf->max_x);
				else
				   index = MAX(0,MIN(i*half_radius+k-box+offset,hf->max_y-1))*hf->max_x
						 + MAX(0,MIN(j*half_radius +l-box+offset,hf->max_x-1)) ;
				indexf = VECTORIZE(CENTER_XY(k,filter->hf_size),
						CENTER_XY(l,filter->hf_size),filter->hf_size>>1) ;
// printf("INDEX dans filter->value: %d\n", indexf );
				*(hf_values + index) = *(hf_values + index) +
					( (gfloat) *(blocks+i*nblocks+j) ) * *(filter->values+ indexf) ;
				*(hf_weight+index) = *(hf_weight+index) +  *(filter->values+ indexf);
// printf("(i,j,k,l): (%d,%d,%d,%d); INDEX: %d; values: %f; w: %f\n",i,j,k,l,index,*(hf_values + index) ,*(hf_weight+index) );
			}
	  }
	} 

	for (i=0; i<hf->max_x*hf->max_y; i++) {
// printf("I: %d; hf_values: %f; hf_weight: %f\n",i,*(hf_values+i),*(hf_weight+i) );
		*(hf->hf_buf+i) = (hf_type) ( (*(hf_values+i))/(*(hf_weight+i)) )  ;
	}
	free(blocks);
	free(hf_values);
	free(hf_weight);
	filter_free(filter);
//	We reapply the process for smoothing the artifacts of the rebuilt HF
	hf_smooth(hf,36,dm, if_tiles, gauss_list);
	}	// End else radius <= 36
}

void hf_mosaic (hf_struct_type *hf, gint radius, gint nblocks, gfloat phase) {
	gint i,j;
	hf_type *blocks;
//	Build the mosaic
	blocks = init_blocks(hf->tmp_buf, hf->max_x, hf->max_y, radius, nblocks, phase, FALSE);
//	Transfer the mosaic into the hf buffer
	for (i=0; i<hf->max_y; i++)
		for (j=0; j<hf->max_x; j++)
			*(hf->hf_buf+WRAP(i+(gint) (radius*phase),hf->max_x)*hf->max_x+j+
				WRAP((gint) (radius*phase),hf->max_y)) = 
				*(blocks+((gint) (i/radius) )*nblocks + (gint) (j/radius));
}

void hf_honeycomb (hf_type *hf_in, hf_type *hf_out, gint max, gint radius, gint border) {
	// Quantize with an hexagonal structure
	// Distance between cells centers = 2*radius
	// "border" = with of the line separating each hexagon, in pixels
	// This transformation does not preserve tiling
	// Important:  radius must be even, otherwise we get black pixels (rounding problem)
	gint x,y,i,j, nbcellsh, nbcellsw, h, w, index, nrow, npix;
	gint baseh, basew, current_cell, block_h, xoffset, xstep;
	static gfloat tan60=1.73205, sin60=0.86602;
	unsigned long int sum; 
	hf_type *cell_values;
	// Hexagons are not symmetrical in a square world...
	// I chose to put the longer axis on the width
	// h = height of each cell
	// w = width
/*	 _    _ 
	/  \_/  \
	\_/  \_/
*/
	radius = 2* (gint) ceil((gdouble) radius/2.0);
	h = (gint) ((gfloat) radius * sin60);
	w = 3 * radius;
	nbcellsh = (gint) ceil ( ((gdouble) 2*radius + max) / (gdouble) h);
	nbcellsw = (gint) ceil (((gdouble) 2*radius + max) / (gdouble) w);
	cell_values = (hf_type *) calloc(1+nbcellsw*nbcellsh,sizeof(hf_type));
	// cell_values contains the average value for the current cell
	// We use hf->hf_buf to store the index of the cell to which the related pixel belongs
	for (i=0; i<max*max; i++)
		*(hf_out+i) = 0;
	// Processing the neighbourhood of each cell...
	// We "center" the structure
	baseh = (max%h)/2;
//	basew = (max%radius)/2;
	basew = 0;
	current_cell = 1;
	// Cell 0 value is 0, so we can detect glitches...
	block_h = radius*sin60;
	nrow=0;
// printf("RADIUS: %d; CH: %d; CW: %d; WxH: %d; BASEH: %d; BASEW: %d; BLOCK_h: %d\n",radius, nbcellsh, nbcellsw, nbcellsh*nbcellsw, baseh, basew, block_h);
	for (y=baseh-h/2; y<max+h; y+=h) {
		xstep = 3*radius ;
		// The x offset is not the same depending if the row is even or odd
		if (nrow%2) {
			xoffset = basew-radius/2;  // This is why radius must be even
		}
		else {
			xoffset = basew+radius;
		}
// printf("ROW: %d; XOFFSET: %d; XSTEP: %d; CELL: %d\n",nrow,xoffset,xstep, current_cell);
		nrow++;
		for (x=xoffset; x<max+radius; x+=xstep) {
// printf("X: %d; ",x);
			sum = 0;
			npix = 0;
			for (i=-block_h+border; i<block_h-border; i++) {
				if ((y+i)<0)
					continue;
				if ((y+i)>=max)
					break;
				for (j=-radius+border; j<radius-border; j++) {
					index = x + j;
					if (index<0)
						continue;
					if (index>max)
						break;
					index = VECTORIZE(index,y+i,max);
				//	Test: is it inside the hexagon?
				//	If abs(x) (x is j) <=cos60*radius (<= radius/2), it's certainly OK
				//	otherwise, we must check if abs(i) / (radius-abs(x)) < tan60
					if ( (ABS(j)<= (radius-border)/2) ||
						(( ((gfloat) ABS(i)) / (gfloat) (radius-border-ABS(j)) ) < tan60 ))
					{
						*(hf_out+index) = current_cell;
						npix++;
						sum = sum + *(hf_in+index);
					}
				}
			}
// printf("Current_CELL: %d; NPIX: %d; SUM: %d\n",current_cell,npix,sum);
			if (npix) // Some cells are outside the image, for some radius
				*(cell_values+current_cell) = sum / npix;
			current_cell++;	
		}
	}
	for (i=0; i<max*max; i++)
		if ((nbcellsh*nbcellsw)<*(hf_out+i)) // "Debugging" test - it's not supposed to happen...
			*(hf_out+i) = 0xFFFF;
		else
			*(hf_out+i) = *(cell_values+*(hf_out+i));
	free(cell_values);
}

void hf_hexagon (hf_struct_type *hf, gint radius, gint border, gint mix,
	gint merge_op, gint source_offset, gint result_offset, 
	gint smooth_radius, 
	dist_matrix_struct *dist_matrix,
	gboolean apply_postprocess,
	gboolean merge_change,
	gint merge_x_translate,
	gint merge_y_translate,
	gdouble **gauss_list) {

	gint i, index, index_offset;
	hf_type hf_source, hf_result;
	gfloat source, result, s_offset, r_offset;

	if (radius<4)
		return;
	
	if (!hf->tmp_buf)
		hf_backup(hf);
	// We need a second temporary buffer for storing
	// the result of the quantization process before the postprocessing
	if (!hf->result_buf)
		hf->result_buf = (hf_type *)malloc(hf->max_x*hf->max_y*sizeof(hf_type));

	if (!merge_change)
		hf_honeycomb (hf->tmp_buf, hf->result_buf, hf->max_x, radius, border);

   	if ( apply_postprocess && smooth_radius  ) {
		// Smooth works on hf_buf, so we must swap it with result_buf
		swap_buffers ((gpointer*) &hf->result_buf, (gpointer*) &hf->hf_buf);
		if (smooth_radius) 
			hf_smooth(hf, smooth_radius, dist_matrix, FALSE, gauss_list);
		swap_buffers ((gpointer*) &hf->result_buf, (gpointer*) &hf->hf_buf);
  	}	// 	end of "if (apply_postprocess)"
 
/*	// Skip the merge when the user asks for the result or the source at 100 % level
	// (commented until more consistency thinking & testing)
	if (mix == 100)
		memcpy(hf->hf_buf,hf->result_buf, hf->max_x*hf->max_y*sizeof(hf_type));
	else
		if (mix == -100)
			memcpy(hf->hf_buf,hf->tmp_buf, hf->max_x*hf->max_y*sizeof(hf_type));
		else
*/
		/*
			// source_offset and result_offset are % of max value >>1
			s_offset = ((MAX_HF_VALUE>>1) * source_offset) / 100.0;	
			r_offset = ((MAX_HF_VALUE>>1) * result_offset) / 100.0;
		*/
			// Changed 2005-06-23: now the offsets are relative values, in %
			// The brightness increase is multiplicative instead of additive (more like contrast)
	 		hf_simple_merge_translate (hf->tmp_buf, hf->result_buf, 
				hf->hf_buf, mix, source_offset, result_offset, 
				hf->max_x, hf->max_y, merge_op, 
				merge_x_translate, merge_y_translate);		
}


void hf_city(hf_struct_type *hf, gint radius, gint skyscraper_var, gint streets_width, merge_struct *mrg) {
//	radius:  skyscraper width, in pixels
//	skyscaper_var:  upper limit of random width variation, in % of radius
//	streets_width:  width of black lines between blocks simulating streets, in % of radius
	gint i,j,k,l,nblocks,str_w, var;
	hf_type *blocks;
	gint *rand_y, *rand_x, s_offset, r_offset;
	
	if (!radius)
		return;
		
	if (!hf->tmp_buf)
		hf_backup(hf);
	if (!hf->result_buf)
		hf->result_buf = (hf_type *)malloc(hf->max_x*hf->max_y*sizeof(hf_type));
	
	if (!mrg->change) {	
		nblocks = (gint) ((hf->max_x-1+radius) / radius);
	//	Sample the height of skyscapers
		blocks = init_blocks(hf->tmp_buf, hf->max_x, hf->max_y, radius, nblocks, 0, FALSE);
	//	Transfer the "mosaic" into the hf buffer
	//	Apply random width variation, bounded by skyscraper_var %, and draw streets
		str_w = (streets_width*radius)/200;	// Divide by 200:  a street extends one half on each block
		rand_x = (gint *) malloc((nblocks+2)*sizeof(gint));	
		rand_y = (gint *) malloc((nblocks+2)*sizeof(gint));
		var = (skyscraper_var*radius)/100;
		*(rand_x) = 0;
		*(rand_y) = 0;
		if (var) {
			for (i=1; i<(nblocks+1); i++) {
				*(rand_y+i) = ((i-1)*radius) + ((rand()%var) - (var/2));
				*(rand_x+i) = ((i-1)*radius) + ((rand()%var) - (var/2));
	// printf("RAND_X(%d): %d;  RAND_Y(%d): %d\n",i, *(rand_x+i),i,*(rand_y+i));
			}
		}
		else
			for (i=1; i<(nblocks+1); i++) {
				*(rand_y+i) = (i-1)*radius;
				*(rand_x+i) = (i-1)*radius;
			}
		*(rand_y+nblocks+1) = hf->max_y;
		*(rand_x+nblocks+1) = hf->max_x;
		for (i=0; i<(nblocks+1); i++) {
			for (j=0; j<(nblocks+1); j++) {
				for (k= *(rand_y+i); (k<*(rand_y+i+1)) ; k++)
					for (l= *(rand_x+j); (l<*(rand_x+j+1)) ; l++)
						if ((k<(*(rand_y+i) + str_w)) || (k>*(rand_y+i+1)-str_w)   || 
							(l<(*(rand_x+j) + str_w)) || (l>*(rand_x+j+1)-str_w) )
							*(hf->result_buf+WRAP(k,hf->max_y)*hf->max_x+
								WRAP(l,hf->max_x)) = 0;
						else
							*(hf->result_buf+WRAP(k,hf->max_y)*hf->max_x+
							WRAP(l,hf->max_x)) = 
							*(blocks+WRAP(i-1,nblocks)*nblocks + WRAP(j-1,nblocks)) ;
			}
		}
		if (rand_y)
			free(rand_y);
		if (rand_x)
			free(rand_x);
	}
/*	
	// source_offset and result_offset are % of max value >>1
	s_offset = ((MAX_HF_VALUE>>1) * mrg->source_offset) / 100.0;	
	r_offset = ((MAX_HF_VALUE>>1) * mrg->result_offset) / 100.0;
*/
	// Changed 2005-06-23: now the offsets are relative values, in %
	// The brightness increase is multiplicative instead of additive (more like contrast)
	
	hf_simple_merge_translate (hf->tmp_buf, hf->result_buf, hf->hf_buf,
			mrg->mix,
			mrg->source_offset, // s_offset,
			mrg->result_offset, // r_offset,
			hf->max_x,
			hf->max_y,
			mrg->merge_op,
			mrg->x_translate,
			mrg->y_translate);
	
/*	Simple version:  skyscraper width is constant
	for (i=0; i<hf->max_y; i++) {
		for (j=0; j<hf->max_x; j++) {
			if ( ( ( (i+(str_w>>1) )%radius) < str_w) || ( ( (j+(str_w>>1))%radius) < str_w) )
				*(hf->hf_buf+i*hf->max_x+j) = 0;
			else
				*(hf->hf_buf+i*hf->max_x+j) =
					*(blocks+((gint) (i/radius) )*nblocks + (gint) (j/radius));
		}
	}
*/
}

hf_type *hf_subtract(hf_type *hf1, hf_type *hf2, gint length) {
//	Subtracts hf2 from hf1, returns a newly allocated hf buffer
	gint i,diff,min=0;
	hf_type *answer;
	answer = (hf_type *) malloc(sizeof(hf_type) * length);
//	2-pass:  1st we find the minimum, 2nd we calculate the adjusted diff. so that it is >0
	for (i=0; i<length; i++) {
		diff = (gint) *(hf1+i)- (gint) *(hf2+i);
		if (diff<min)
			min = diff;
	}
	for (i=0; i<length; i++)
		*(answer+i) =  (hf_type) ((gint) *(hf1+i) - (gint) *(hf2+i) - min);
	return answer;
}

void hf_stretch_v (hf_struct_type *hf) {
//	Stretch vertically (compress horizontally)
//	We simply divide the scale by 2 horizontally
	gint x,y;
	if (!hf->tmp_buf)
		hf_backup(hf);
	for (y=0; y<hf->max_y; y++) {
		for (x=0; x<hf->max_x; x++) {
			*(hf->hf_buf + VECTORIZE(x,y,hf->max_x)) =
				*(hf->tmp_buf + VECTORIZE(WRAP(x<<1,hf->max_x),y,hf->max_x)); 
		}
	}
}
void hf_stretch_h (hf_struct_type *hf) {
//	Stretch horizontally (compress vertically)
//	We simply divide the scale by 2 vertically
	gint x,y;
	if (!hf->tmp_buf)
		hf_backup(hf);
	for (y=0; y<hf->max_y; y++) {
		for (x=0; x<hf->max_x; x++) {
			*(hf->hf_buf + VECTORIZE(x,y,hf->max_x)) =
				*(hf->tmp_buf + VECTORIZE(x,WRAP((y<<1),hf->max_y),hf->max_x)); 
		}
	}
}

void hf_math_fn (hf_struct_type *hf, gint math_op, gdouble parameter) {
//	Some mathematical pixel transformations
	gint i;
	static gdouble PI2 = 6.283185307, maxd = (gdouble) 0xFFFF;
	gdouble *doublebuf, ratiolog;
	doublebuf = alloc_double_hf(hf->max_x);
	ratiolog = (10.0-parameter) * log1p(maxd) / maxd;
// printf("MAXD:  %f; LOG(MAX): %f;  RATIOLOG: %f\n",maxd, log1p(maxd), ratiolog);
	if (!hf->tmp_buf)
		hf_backup(hf);
// printf("MATH_OP: %d; PARAM: %f\n",math_op, parameter);

	for (i=0; i<hf->max_x*hf->max_y; i++) {

		switch (math_op) {
		case POWER_OP:
			*(doublebuf +i) = pow( (gdouble) *(hf->tmp_buf+i), parameter);
			break;
		case BASE_OP:
			*(doublebuf +i) = pow( parameter, ((gdouble) *(hf->tmp_buf+i)) / maxd );
			break;
		case LOG_OP:
			*(doublebuf+i) = ratiolog * *(hf->tmp_buf+i) + parameter * log1p((gdouble)*(hf->tmp_buf+i));
			break;
		case SINE_OP:
			*(doublebuf+i) = sin( parameter * PI2 *((gdouble) *(hf->tmp_buf+i)) / maxd );
		}	
	}
	hf_min_max(hf);
	hf_double_clamp (hf, hf->min, hf->max, doublebuf);
	free(doublebuf);
}
